/* 
 * Copyright (C) 2017 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/gpl-3.0.txt
 *
 */
package com.frojasg1.chesspdfbrowser.application;

import com.frojasg1.chesspdfbrowser.configuration.AppStringsConf;
import com.frojasg1.chesspdfbrowser.configuration.ApplicationConfiguration;
import com.frojasg1.chesspdfbrowser.configuration.ChessStringsConf;
import com.frojasg1.chesspdfbrowser.exception.ChessGamePositionException;
import com.frojasg1.chesspdfbrowser.exception.ChessModelException;
import com.frojasg1.chesspdfbrowser.exception.ChessMoveException;
import com.frojasg1.chesspdfbrowser.exception.ChessPieceCreationException;
import com.frojasg1.chesspdfbrowser.exception.ChessWriterException;
import com.frojasg1.chesspdfbrowser.io.file.implementation.PgnChessFile;
import com.frojasg1.chesspdfbrowser.io.parsers.implementation.PGNChessGameParser;
import com.frojasg1.chesspdfbrowser.io.parsers.implementation.RawChessGameParser;
import com.frojasg1.chesspdfbrowser.model.chess.board.ChessGamePosition;
import com.frojasg1.chesspdfbrowser.model.chess.configuration.EnglishChessConfiguration;
import com.frojasg1.chesspdfbrowser.model.chess.game.ChessGame;
import com.frojasg1.chesspdfbrowser.model.chess.game.ChessGameHeaderInfo;
import com.frojasg1.chesspdfbrowser.model.chess.game.tree.ChessGameMove;
import com.frojasg1.chesspdfbrowser.model.chess.game.tree.MoveTreeNode;
import com.frojasg1.chesspdfbrowser.threads.LoadControllerInterface;
import com.frojasg1.chesspdfbrowser.threads.LoadPDFThread;
import com.frojasg1.chesspdfbrowser.threads.LoadPGNThread;
import com.frojasg1.chesspdfbrowser.threads.SavePGNThread;
import com.frojasg1.chesspdfbrowser.view.chess.ChessBoardPanel;
import com.frojasg1.chesspdfbrowser.view.chess.ChessGameListTable;
import com.frojasg1.chesspdfbrowser.view.chess.ChessTreeGameTextPane;
import com.frojasg1.chesspdfbrowser.view.chess.editcomment.EditCommentFrame;
import com.frojasg1.general.desktop.view.panels.NavigatorControllerInterface;
import com.frojasg1.general.desktop.view.panels.NavigatorJPanel;
import com.frojasg1.chesspdfbrowser.view.chess.edittags.EditTAGsJFrame;
import com.frojasg1.chesspdfbrowser.view.chess.initialposition.InitialPositionDialog;
import com.frojasg1.chesspdfbrowser.view.chess.interaction.ChessGameControllerInterface;
import com.frojasg1.chesspdfbrowser.view.chess.multiwindowmanager.DetachedGameWindow;
import com.hotmail.frojasg1.applications.common.components.internationalization.window.InternationalizedJFrame;
import com.hotmail.frojasg1.applications.common.components.resizecomp.MapResizeRelocateComponentItem;
import com.hotmail.frojasg1.applications.common.components.resizecomp.ResizeRelocateItem;
import com.frojasg1.general.exceptions.ConfigurationException;
import com.frojasg1.general.dialogs.DialogsWrapper;
import java.awt.Component;
import java.awt.Container;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.io.BufferedOutputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.List;
import javax.swing.JMenu;
import org.apache.pdfbox.pdmodel.PDDocument;
import com.frojasg1.chesspdfbrowser.view.chess.multiwindowmanager.MultiwindowGameManager;
import com.frojasg1.general.desktop.clipboard.SystemClipboard;
import com.frojasg1.general.desktop.files.FileFunctions;
import com.frojasg1.general.desktop.view.ToolTipLookAndFeel;
import com.frojasg1.general.desktop.view.about.GenericAboutJDialog;
import com.frojasg1.general.desktop.view.panels.InformerInterface;
import com.frojasg1.general.dialogs.FilterForFile;
import com.frojasg1.general.number.IntegerFunctions;
import com.frojasg1.generic.GenericFunctions;
import java.awt.Frame;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Vector;
import javax.swing.JPopupMenu;

/**
 *
 * @author Usuario
 */
public class MainWindow extends InternationalizedJFrame  implements ChessGameControllerInterface, NavigatorControllerInterface,
																					LoadControllerInterface, KeyListener
{
	public static final String sa_configurationBaseFileName = "MainWindow";

	protected static final String CONF_CURRENT_LIST_OF_GAMES_HAS_BEEN_MODIFIED_DO_YOU_WANT_TO_SAVE_CHANGES = "CURRENT_LIST_OF_GAMES_HAS_BEEN_MODIFIED_DO_YOU_WANT_TO_SAVE_CHANGES";
	protected static final String CONF_WARNING = "WARNING";
	protected static final String CONF_YES = "YES";
	protected static final String CONF_NO = "NO";
	protected static final String CONF_CANCEL = "CANCEL";

//	protected String _language = null;
//	protected String _additionalLanguage = null;

	protected ChessGame _currentChessGame = null;

	protected List<ChessGame> _listOfGames = null;
	protected boolean _listOfGamesHasBeenModified = false;

	protected PgnChessFile _pgnFile = null;

	protected ChessTreeGameTextPane jTextPane1;
	protected ChessGameListTable jTListOfGames = null;
	protected ChessBoardPanel _chessBoardPanel = null;
	protected NavigatorJPanel _navigatorPanel = null;

	protected PDFviewerWindow _pdfViewerWindow = null;

	protected String _lastFileName = null;

	protected MultiwindowGameManager _multiWindowGameManager = null;
	protected MainWindow _this = null;

	protected AppStringsConf _appStrConf = null;

	protected Map< Frame, Integer > _stateOfWindows = null;

//	protected EditCommentFrame _editCommentFrame = null;
//	protected EditTAGsJFrame

	/**
	 * Creates new form ChessGameBrowserWindow
	 */
	public MainWindow( ) throws ConfigurationException
	{
		super( ApplicationConfiguration.instance() );

		initConfiguration();

		initComponents();

		initOwnComponents();

		setWindowConfiguration();

		setKeyListener( this );

		_this = this;

		createExternalWindows();

		Splash.instance().setProgress( 100 );
		
		try
		{
			Thread.sleep(200);
		}
		catch( InterruptedException ie )
		{
			ie.printStackTrace();
		}
//		applyNewApplicationConfiguration();
	}

	protected void initConfiguration() throws ConfigurationException
	{
/*
		try
		{
			getAppliConf().M_openConfiguration();
		}
		catch( ConfigurationException ce )
		{
			ce.printStackTrace();

			DialogsWrapper.showMessageDialog(null, ce.getMessage() + " Exiting from application",
											"Configuration error", JOptionPane.ERROR_MESSAGE, null );
			throw( ce );
		}

		try
		{
			_appStrConf = AppStringsConf.createInstance( ApplicationConfiguration.sa_PROPERTIES_PATH_IN_JAR,
														getAppliConf() );
			_appStrConf.M_openConfiguration();
//			_appStrConf.changeLanguage( getAppliConf().M_getLanguage() );
		}
		catch( Throwable th )
		{
			th.printStackTrace();
		}
*/
		_appStrConf = AppStringsConf.instance();

/*
		ChessStringsConf chessStrConf = null;
		try
		{
			chessStrConf = ChessStringsConf.createInstance( ApplicationConfiguration.sa_PROPERTIES_PATH_IN_JAR,
															getAppliConf() );
			chessStrConf.M_openConfiguration();
//			chessStrConf.changeLanguage( getAppliConf().M_getLanguage() );
		}
		catch( Throwable th )
		{
			th.printStackTrace();
		}
*/
	}

	protected void createExternalWindows()
	{
		java.awt.EventQueue.invokeLater(new Runnable()
		{
			public void run()
			{
				try
				{
					InitialPositionDialog.createInstance(_this, _appliConf);
					EditCommentFrame.createInstance( _this, getAppliConf(), _this );
					EditTAGsJFrame.createInstance( _this, getAppliConf(), _this );

					setInitialGame();
					applyNewApplicationConfiguration();
				}
				catch( Throwable th )
				{
					th.printStackTrace();
				}
			}
		});
	}

	@Override
	public ApplicationConfiguration getAppliConf()
	{
		ApplicationConfiguration result = null;
		if( _appliConf instanceof ApplicationConfiguration )
			result = (ApplicationConfiguration) _appliConf;

		return( result );
	}

	protected void setInitialGame()
	{
		try
		{
			ChessGame cg = new ChessGame( getAppliConf() );

			List<ChessGame> listOfGames = new ArrayList<ChessGame>();
			listOfGames.add( cg );

			changeListOfChessGames(listOfGames, 0);
		}
		catch( Throwable th )
		{
			th.printStackTrace();
		}
	}

	protected void initOwnComponents()
	{
		try
		{
//			RBG_language.add(jRadioButtonMenuItemIdiomaEn);
//			RBG_language.add(jRadioButtonMenuItemIdiomaEs);
//			RBG_language.add(jRadioButtonMenuItemIdiomaOtro);
			
			_chessBoardPanel = new ChessBoardPanel();
			_chessBoardPanel.setName( "chessBoardPanel" );
//			_currentChessGame = new ChessGame( getAppliConf() );

//			_chessBoardPanel.setChessBoard( _currentChessGame.getChessBoard() );
			_chessBoardPanel.setChessBoard( null );

			jPanelChessBoardContainer.add( _chessBoardPanel );
			_chessBoardPanel.setBounds( 0, 0, jPanelChessBoardContainer.getWidth(), jPanelChessBoardContainer.getHeight() );

			_chessBoardPanel.setListener( this );

			_multiWindowGameManager = new MultiwindowGameManager( this, getAppliConf() );

			jTextPane1 = new ChessTreeGameTextPane(_chessBoardPanel, this, _multiWindowGameManager, true);
			jTextPane1.setChessGame(null);

			_multiWindowGameManager.setParentChessTreeGameTextPane(jTextPane1);

			jScrollPane2.setViewportView(jTextPane1);

			jTListOfGames = new ChessGameListTable( this, getAppliConf(), _multiWindowGameManager );
			jScrollPane1.setViewportView(jTListOfGames);

			jTListOfGames.initialize();
			
			_navigatorPanel = new NavigatorJPanel( this );
			_navigatorPanel.setBounds(0, 0, 230, 50);
			jPanel3.add( _navigatorPanel );

			pack();
		}
		catch( Throwable th )
		{
			th.printStackTrace();
		}
	}

	protected void setWindowConfiguration( )
	{
		Vector<JPopupMenu> vectorJpopupMenus = new Vector<JPopupMenu>();
		vectorJpopupMenus.add( jTextPane1.getJPopupMenu() );
		vectorJpopupMenus.add( jTListOfGames.getJPopupMenu() );

		MapResizeRelocateComponentItem mapRRCI = new MapResizeRelocateComponentItem();
		try
		{
			mapRRCI.putResizeRelocateComponentItem( jPanel1, ResizeRelocateItem.RESIZE_TO_RIGHT + ResizeRelocateItem.RESIZE_TO_BOTTOM );
			mapRRCI.putResizeRelocateComponentItem( jSPmainSplit, ResizeRelocateItem.RESIZE_TO_RIGHT + ResizeRelocateItem.RESIZE_TO_BOTTOM );
			mapRRCI.putResizeRelocateComponentItem( jPanelNavigatorContainer1, ResizeRelocateItem.RESIZE_TO_RIGHT + ResizeRelocateItem.RESIZE_TO_BOTTOM );
			mapRRCI.putResizeRelocateComponentItem( jSPnavigatorSplit, ResizeRelocateItem.RESIZE_TO_RIGHT + ResizeRelocateItem.RESIZE_TO_BOTTOM );
			mapRRCI.putResizeRelocateComponentItem( jPanel4, ResizeRelocateItem.RESIZE_TO_RIGHT );
			mapRRCI.putResizeRelocateComponentItem( jPanel3, ResizeRelocateItem.RESIZE_TO_RIGHT );
			mapRRCI.putResizeRelocateComponentItem( _navigatorPanel, ResizeRelocateItem.RESIZE_TO_RIGHT + ResizeRelocateItem.RESIZE_TO_BOTTOM );

			mapRRCI.putResizeRelocateComponentItem( jScrollPane2, ResizeRelocateItem.RESIZE_TO_RIGHT + ResizeRelocateItem.RESIZE_TO_BOTTOM );
			mapRRCI.putResizeRelocateComponentItem( jTextPane1, ResizeRelocateItem.RESIZE_TO_RIGHT + ResizeRelocateItem.RESIZE_TO_BOTTOM );
			mapRRCI.putResizeRelocateComponentItem( jScrollPane1, ResizeRelocateItem.RESIZE_TO_RIGHT + ResizeRelocateItem.RESIZE_TO_BOTTOM );
			mapRRCI.putResizeRelocateComponentItem( jTListOfGames, ResizeRelocateItem.RESIZE_TO_RIGHT + ResizeRelocateItem.RESIZE_TO_BOTTOM );

			mapRRCI.putResizeRelocateComponentItem( jPanelChessBoardContainer, ResizeRelocateItem.RESIZE_TO_RIGHT + ResizeRelocateItem.RESIZE_TO_BOTTOM );
			mapRRCI.putResizeRelocateComponentItem( _chessBoardPanel, ResizeRelocateItem.RESIZE_TO_RIGHT + ResizeRelocateItem.RESIZE_TO_BOTTOM );
		}
		catch( Throwable th )
		{
			th.printStackTrace();
		}
/*
		a_intern = new JFrameInternationalization(	ApplicationConfiguration.sa_MAIN_FOLDER,
													ApplicationConfiguration.sa_APPLICATION_NAME,
													ApplicationConfiguration.sa_CONFIGURATION_GROUP,
													ApplicationConfiguration.sa_PROPERTIES_PATH_IN_JAR,
													a_configurationBaseFileName,
													this,
													parent,
													a_vectorJpopupMenus,
													false,
													mapRRCI );
*/
		createInternationalization(	getAppliConf().getConfigurationMainFolder(),
									getAppliConf().getApplicationNameFolder(),
									getAppliConf().getApplicationGroup(),
									getAppliConf().getInternationalPropertiesPathInJar(),
									sa_configurationBaseFileName,
									this,
									null,
									vectorJpopupMenus,
									true,
									mapRRCI );

		jTextPane1.setLanguageConfiguration( this );
		jTListOfGames.setLanguageConfiguration( this );

		this.registerInternationalString(CONF_CURRENT_LIST_OF_GAMES_HAS_BEEN_MODIFIED_DO_YOU_WANT_TO_SAVE_CHANGES,
											"Current list of games has been modified since last saved. Do you want to save it now?" );
		this.registerInternationalString(CONF_WARNING, "Warning" );
		this.registerInternationalString(CONF_YES, "Yes" );
		this.registerInternationalString(CONF_NO, "No" );
		this.registerInternationalString(CONF_CANCEL, "Cancel" );

		setLanguageSubmenu( jM_languageSubmenu.getPopupMenu() );

//		setIcon( "com/frojasg1/chesspdfbrowser/resources/icons/App.icon.png" );
	}

	/**
	 * This method is called from within the constructor to initialize the form.
	 * WARNING: Do NOT modify this code. The content of this method is always
	 * regenerated by the Form Editor.
	 */
	@SuppressWarnings("unchecked")
    // <editor-fold defaultstate="collapsed" desc="Generated Code">//GEN-BEGIN:initComponents
    private void initComponents() {

        jPanel1 = new javax.swing.JPanel();
        jSPmainSplit = new javax.swing.JSplitPane();
        jSPnavigatorSplit = new javax.swing.JSplitPane();
        jPanelNavigatorContainer1 = new javax.swing.JPanel();
        jPanel4 = new javax.swing.JPanel();
        jCB_whitePlaysFromBottom = new javax.swing.JCheckBox();
        jSeparator3 = new javax.swing.JSeparator();
        jPanel3 = new javax.swing.JPanel();
        jScrollPane2 = new javax.swing.JScrollPane();
        jScrollPane1 = new javax.swing.JScrollPane();
        jPanelChessBoardContainer = new javax.swing.JPanel();
        jMenuBar1 = new javax.swing.JMenuBar();
        jM_File = new javax.swing.JMenu();
        jMI_open = new javax.swing.JMenuItem();
        jSeparator4 = new javax.swing.JPopupMenu.Separator();
        jMI_save = new javax.swing.JMenuItem();
        jMI_saveAs = new javax.swing.JMenuItem();
        jSeparator2 = new javax.swing.JPopupMenu.Separator();
        jMI_exit = new javax.swing.JMenuItem();
        jM_Edit = new javax.swing.JMenu();
        jMI_copyFEN = new javax.swing.JMenuItem();
        jMI_copyPGN = new javax.swing.JMenuItem();
        jSeparator1 = new javax.swing.JPopupMenu.Separator();
        jMI_paste = new javax.swing.JMenuItem();
        jM_preferences = new javax.swing.JMenu();
        jM_languageSubmenu = new javax.swing.JMenu();
        jRadioButtonMenuItemIdiomaEn = new javax.swing.JRadioButtonMenuItem();
        jRadioButtonMenuItemIdiomaEs = new javax.swing.JRadioButtonMenuItem();
        jRadioButtonMenuItemIdiomaOtro = new javax.swing.JRadioButtonMenuItem();
        jSeparator9 = new javax.swing.JPopupMenu.Separator();
        jMI_preferences = new javax.swing.JMenuItem();
        jM_windows = new javax.swing.JMenu();
        jMI_editComment = new javax.swing.JMenuItem();
        jSeparator7 = new javax.swing.JPopupMenu.Separator();
        jMI_editTAGsWindow = new javax.swing.JMenuItem();
        jSeparator6 = new javax.swing.JPopupMenu.Separator();
        jMI_editInitialPosition = new javax.swing.JMenuItem();
        jSeparator10 = new javax.swing.JPopupMenu.Separator();
        jMI_openCurrentGameInDetachedWindow = new javax.swing.JMenuItem();
        jM_about = new javax.swing.JMenu();
        jMI_about = new javax.swing.JMenuItem();
        jSeparator5 = new javax.swing.JPopupMenu.Separator();
        jMI_license = new javax.swing.JMenuItem();

        setDefaultCloseOperation(javax.swing.WindowConstants.DO_NOTHING_ON_CLOSE);
        setMinimumSize(new java.awt.Dimension(600, 415));
        setName(""); // NOI18N
        addWindowListener(new java.awt.event.WindowAdapter() {
            public void windowClosing(java.awt.event.WindowEvent evt) {
                formWindowClosing(evt);
            }
            public void windowDeiconified(java.awt.event.WindowEvent evt) {
                formWindowDeiconified(evt);
            }
            public void windowIconified(java.awt.event.WindowEvent evt) {
                formWindowIconified(evt);
            }
        });

        jPanel1.setMinimumSize(new java.awt.Dimension(584, 270));
        jPanel1.setPreferredSize(new java.awt.Dimension(600, 385));

        jSPmainSplit.setDividerLocation(233);
        jSPmainSplit.setResizeWeight(0.2);
        jSPmainSplit.setContinuousLayout(true);
        jSPmainSplit.setName("jSPmainSplit"); // NOI18N

        jSPnavigatorSplit.setDividerLocation(240);
        jSPnavigatorSplit.setOrientation(javax.swing.JSplitPane.VERTICAL_SPLIT);
        jSPnavigatorSplit.setResizeWeight(0.8);
        jSPnavigatorSplit.setContinuousLayout(true);
        jSPnavigatorSplit.setName("jSPnavigatorSplit"); // NOI18N

        jPanelNavigatorContainer1.setMinimumSize(new java.awt.Dimension(230, 190));
        jPanelNavigatorContainer1.setLayout(null);

        jPanel4.setBorder(javax.swing.BorderFactory.createBevelBorder(javax.swing.border.BevelBorder.RAISED));
        jPanel4.setMinimumSize(new java.awt.Dimension(230, 70));
        jPanel4.setLayout(null);

        jCB_whitePlaysFromBottom.setSelected(true);
        jCB_whitePlaysFromBottom.setText("white plays from bottom");
        jCB_whitePlaysFromBottom.setName("jCB_whitePlaysFromBottom"); // NOI18N
        jCB_whitePlaysFromBottom.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                jCB_whitePlaysFromBottomActionPerformed(evt);
            }
        });
        jPanel4.add(jCB_whitePlaysFromBottom);
        jCB_whitePlaysFromBottom.setBounds(10, 10, 190, 23);
        jCB_whitePlaysFromBottom.getAccessibleContext().setAccessibleDescription("Esto es una prueba");

        jPanel4.add(jSeparator3);
        jSeparator3.setBounds(50, 20, 0, 2);

        jPanelNavigatorContainer1.add(jPanel4);
        jPanel4.setBounds(0, 0, 230, 70);

        jPanel3.setBorder(javax.swing.BorderFactory.createBevelBorder(javax.swing.border.BevelBorder.RAISED));
        jPanel3.setMinimumSize(new java.awt.Dimension(230, 50));
        jPanel3.setName(""); // NOI18N
        jPanel3.setLayout(null);
        jPanelNavigatorContainer1.add(jPanel3);
        jPanel3.setBounds(0, 70, 230, 50);

        jScrollPane2.setMinimumSize(new java.awt.Dimension(230, 70));
        jScrollPane2.setPreferredSize(new java.awt.Dimension(230, 120));
        jScrollPane2.setRequestFocusEnabled(false);
        jPanelNavigatorContainer1.add(jScrollPane2);
        jScrollPane2.setBounds(0, 120, 230, 120);

        jSPnavigatorSplit.setLeftComponent(jPanelNavigatorContainer1);

        jScrollPane1.setMinimumSize(new java.awt.Dimension(230, 70));
        jScrollPane1.setPreferredSize(new java.awt.Dimension(230, 230));
        jSPnavigatorSplit.setBottomComponent(jScrollPane1);

        jSPmainSplit.setLeftComponent(jSPnavigatorSplit);

        jPanelChessBoardContainer.setBorder(javax.swing.BorderFactory.createLineBorder(new java.awt.Color(0, 0, 0)));
        jPanelChessBoardContainer.setMinimumSize(new java.awt.Dimension(350, 350));

        javax.swing.GroupLayout jPanelChessBoardContainerLayout = new javax.swing.GroupLayout(jPanelChessBoardContainer);
        jPanelChessBoardContainer.setLayout(jPanelChessBoardContainerLayout);
        jPanelChessBoardContainerLayout.setHorizontalGroup(
            jPanelChessBoardContainerLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addGap(0, 459, Short.MAX_VALUE)
        );
        jPanelChessBoardContainerLayout.setVerticalGroup(
            jPanelChessBoardContainerLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addGap(0, 475, Short.MAX_VALUE)
        );

        jSPmainSplit.setRightComponent(jPanelChessBoardContainer);

        javax.swing.GroupLayout jPanel1Layout = new javax.swing.GroupLayout(jPanel1);
        jPanel1.setLayout(jPanel1Layout);
        jPanel1Layout.setHorizontalGroup(
            jPanel1Layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addComponent(jSPmainSplit)
        );
        jPanel1Layout.setVerticalGroup(
            jPanel1Layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addComponent(jSPmainSplit)
        );

        jM_File.setText("File");
        jM_File.setName("jM_File"); // NOI18N

        jMI_open.setText("Open");
        jMI_open.setName("jMI_open"); // NOI18N
        jMI_open.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                jMI_openActionPerformed(evt);
            }
        });
        jM_File.add(jMI_open);
        jM_File.add(jSeparator4);

        jMI_save.setText("Save");
        jMI_save.setName("jMI_save"); // NOI18N
        jMI_save.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                jMI_saveActionPerformed(evt);
            }
        });
        jM_File.add(jMI_save);

        jMI_saveAs.setText("Save as");
        jMI_saveAs.setName("jMI_saveAs"); // NOI18N
        jMI_saveAs.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                jMI_saveAsActionPerformed(evt);
            }
        });
        jM_File.add(jMI_saveAs);
        jM_File.add(jSeparator2);

        jMI_exit.setText("Exit");
        jMI_exit.setName("jMI_exit"); // NOI18N
        jMI_exit.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                jMI_exitActionPerformed(evt);
            }
        });
        jM_File.add(jMI_exit);

        jMenuBar1.add(jM_File);

        jM_Edit.setText("Edit");
        jM_Edit.setName("jM_Edit"); // NOI18N

        jMI_copyFEN.setText("Copy FEN");
        jMI_copyFEN.setName("jMI_copyFEN"); // NOI18N
        jMI_copyFEN.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                jMI_copyFENActionPerformed(evt);
            }
        });
        jM_Edit.add(jMI_copyFEN);

        jMI_copyPGN.setText("Copy PGN");
        jMI_copyPGN.setName("jMI_copyPGN"); // NOI18N
        jMI_copyPGN.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                jMI_copyPGNActionPerformed(evt);
            }
        });
        jM_Edit.add(jMI_copyPGN);
        jM_Edit.add(jSeparator1);

        jMI_paste.setText("Paste");
        jMI_paste.setName("jMI_paste"); // NOI18N
        jMI_paste.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                jMI_pasteActionPerformed(evt);
            }
        });
        jM_Edit.add(jMI_paste);

        jMenuBar1.add(jM_Edit);

        jM_preferences.setText("Preferences");
        jM_preferences.setName("jM_preferences"); // NOI18N

        jM_languageSubmenu.setText("Language");
        jM_languageSubmenu.setName("jM_languageSubmenu"); // NOI18N

        jRadioButtonMenuItemIdiomaEn.setText("EN");
        jRadioButtonMenuItemIdiomaEn.setName(""); // NOI18N
        jM_languageSubmenu.add(jRadioButtonMenuItemIdiomaEn);

        jRadioButtonMenuItemIdiomaEs.setText("ES");
        jRadioButtonMenuItemIdiomaEs.setName(""); // NOI18N
        jM_languageSubmenu.add(jRadioButtonMenuItemIdiomaEs);

        jRadioButtonMenuItemIdiomaOtro.setText("CAT");
        jRadioButtonMenuItemIdiomaOtro.setName(""); // NOI18N
        jM_languageSubmenu.add(jRadioButtonMenuItemIdiomaOtro);

        jM_preferences.add(jM_languageSubmenu);
        jM_preferences.add(jSeparator9);

        jMI_preferences.setText("Preferences");
        jMI_preferences.setName("jMI_preferences"); // NOI18N
        jMI_preferences.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                jMI_preferencesActionPerformed(evt);
            }
        });
        jM_preferences.add(jMI_preferences);

        jMenuBar1.add(jM_preferences);

        jM_windows.setText("Windows");
        jM_windows.setName("jM_windows"); // NOI18N

        jMI_editComment.setText("Edit comment window");
        jMI_editComment.setName("jMI_editComment"); // NOI18N
        jMI_editComment.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                jMI_editCommentActionPerformed(evt);
            }
        });
        jM_windows.add(jMI_editComment);
        jM_windows.add(jSeparator7);

        jMI_editTAGsWindow.setText("Edit header TAGs window");
        jMI_editTAGsWindow.setName("jMI_editTAGsWindow"); // NOI18N
        jMI_editTAGsWindow.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                jMI_editTAGsWindowActionPerformed(evt);
            }
        });
        jM_windows.add(jMI_editTAGsWindow);
        jM_windows.add(jSeparator6);

        jMI_editInitialPosition.setText("Edit initial position");
        jMI_editInitialPosition.setName("jMI_editInitialPosition"); // NOI18N
        jMI_editInitialPosition.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                jMI_editInitialPositionActionPerformed(evt);
            }
        });
        jM_windows.add(jMI_editInitialPosition);
        jM_windows.add(jSeparator10);

        jMI_openCurrentGameInDetachedWindow.setText("Open current game in detached window");
        jMI_openCurrentGameInDetachedWindow.setName("jMI_openCurrentGameInDetachedWindow"); // NOI18N
        jMI_openCurrentGameInDetachedWindow.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                jMI_openCurrentGameInDetachedWindowActionPerformed(evt);
            }
        });
        jM_windows.add(jMI_openCurrentGameInDetachedWindow);

        jMenuBar1.add(jM_windows);

        jM_about.setText("About");
        jM_about.setName("jM_about"); // NOI18N

        jMI_about.setText("About");
        jMI_about.setName("jMI_about"); // NOI18N
        jMI_about.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                jMI_aboutActionPerformed(evt);
            }
        });
        jM_about.add(jMI_about);
        jM_about.add(jSeparator5);

        jMI_license.setText("License");
        jMI_license.setName("jMI_license"); // NOI18N
        jMI_license.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                jMI_licenseActionPerformed(evt);
            }
        });
        jM_about.add(jMI_license);

        jMenuBar1.add(jM_about);

        setJMenuBar(jMenuBar1);

        javax.swing.GroupLayout layout = new javax.swing.GroupLayout(getContentPane());
        getContentPane().setLayout(layout);
        layout.setHorizontalGroup(
            layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addComponent(jPanel1, javax.swing.GroupLayout.DEFAULT_SIZE, 700, Short.MAX_VALUE)
        );
        layout.setVerticalGroup(
            layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addComponent(jPanel1, javax.swing.GroupLayout.DEFAULT_SIZE, 479, Short.MAX_VALUE)
        );

        pack();
    }// </editor-fold>//GEN-END:initComponents

    private void formWindowClosing(java.awt.event.WindowEvent evt) {//GEN-FIRST:event_formWindowClosing
        // TODO add your handling code here:

		boolean closeWindow = true;

		if( checkToSaveCurrentPGN() )
			formWindowClosing( closeWindow );

    }//GEN-LAST:event_formWindowClosing

    private void jCB_whitePlaysFromBottomActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_jCB_whitePlaysFromBottomActionPerformed
        // TODO add your handling code here:

		_chessBoardPanel.setWhitePlaysFromTheBottom( jCB_whitePlaysFromBottom.isSelected() );
		_chessBoardPanel.repaint();

/*
		LookAndFeel laf = javax.swing.UIManager.getLookAndFeel();
		System.out.println( laf );
		if( laf != null )
			System.out.println( laf.getName() );
*/

//		System.out.println( getSize() );
    }//GEN-LAST:event_jCB_whitePlaysFromBottomActionPerformed

    private void jMI_openActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_jMI_openActionPerformed
        // TODO add your handling code here:

		String fileName = showFileChooserDialog( DialogsWrapper.OPEN );

		if( fileName != null )	openFile( fileName );

    }//GEN-LAST:event_jMI_openActionPerformed

    private void jMI_pasteActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_jMI_pasteActionPerformed
        // TODO add your handling code here:

		parsePastedText();
		
    }//GEN-LAST:event_jMI_pasteActionPerformed

    private void jMI_saveActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_jMI_saveActionPerformed
        // TODO add your handling code here:

		save();
    }//GEN-LAST:event_jMI_saveActionPerformed

    private void jMI_saveAsActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_jMI_saveAsActionPerformed
        // TODO add your handling code here:

		String fileName = showFileChooserDialog( DialogsWrapper.SAVE );

		if( fileName != null )	saveAs( fileName );

    }//GEN-LAST:event_jMI_saveAsActionPerformed

    private void jMI_exitActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_jMI_exitActionPerformed
        // TODO add your handling code here:

		boolean closeWindow = true;
		formWindowClosing( closeWindow );

    }//GEN-LAST:event_jMI_exitActionPerformed

    private void jMI_editCommentActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_jMI_editCommentActionPerformed
        // TODO add your handling code here:

		editComment(_currentChessGame, (_currentChessGame != null ? _currentChessGame.getCurrentMove() : null ) );

    }//GEN-LAST:event_jMI_editCommentActionPerformed

    private void jMI_openCurrentGameInDetachedWindowActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_jMI_openCurrentGameInDetachedWindowActionPerformed
        // TODO add your handling code here:

		DetachedGameWindow dgw = _multiWindowGameManager.addOrFocusGameWindow(_currentChessGame);

    }//GEN-LAST:event_jMI_openCurrentGameInDetachedWindowActionPerformed

    private void jMI_editTAGsWindowActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_jMI_editTAGsWindowActionPerformed
        // TODO add your handling code here:

		EditTAGsJFrame.instance().setVisible( true );
		EditTAGsJFrame.instance().setState ( Frame.NORMAL );

    }//GEN-LAST:event_jMI_editTAGsWindowActionPerformed

    private void jMI_preferencesActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_jMI_preferencesActionPerformed
        // TODO add your handling code here:

		JDial_applicationConfiguration dialog = null;
		try
		{
			dialog = new JDial_applicationConfiguration( this, true );
			dialog.setVisible(true);

			if( dialog.getUserHasPressedOK() )
			{
				applyNewApplicationConfiguration();
			}
		}
		finally
		{
			if( dialog != null )
			{
				dialog.releaseResources();
				dialog.dispose();
				dialog = null;
			}
		}
    }//GEN-LAST:event_jMI_preferencesActionPerformed

    private void jMI_copyPGNActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_jMI_copyPGNActionPerformed
        // TODO add your handling code here:

		if( _currentChessGame != null )
		{
			try
			{
				PgnChessFile pcf = new PgnChessFile();
				String pgnString = pcf.saveGameToFile( _currentChessGame );

				SystemClipboard.instance().setClipboardContents( pgnString );
			}
			catch( Throwable th )
			{
				th.printStackTrace();
			}
		}

    }//GEN-LAST:event_jMI_copyPGNActionPerformed

    private void jMI_copyFENActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_jMI_copyFENActionPerformed
        // TODO add your handling code here:

		if( _currentChessGame != null )
		{
			try
			{
				String pgnString = _currentChessGame.getChessBoard().getCurrentPosition().getFenString();

				SystemClipboard.instance().setClipboardContents( pgnString );
			}
			catch( Throwable th )
			{
				th.printStackTrace();
			}
		}

    }//GEN-LAST:event_jMI_copyFENActionPerformed

    private void jMI_editInitialPositionActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_jMI_editInitialPositionActionPerformed
        // TODO add your handling code here:

		editInitialPosition(_currentChessGame);

    }//GEN-LAST:event_jMI_editInitialPositionActionPerformed

    private void jMI_aboutActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_jMI_aboutActionPerformed
        // TODO add your handling code here:

		GenericAboutJDialog aboutDial = new AboutJDialog( this, true, getAppliConf() );

		aboutDial.setVisible( true );
    }//GEN-LAST:event_jMI_aboutActionPerformed

    private void formWindowIconified(java.awt.event.WindowEvent evt) {//GEN-FIRST:event_formWindowIconified
        // TODO add your handling code here:

		saveStateOfOpenWindowsAndIconifyThem();

    }//GEN-LAST:event_formWindowIconified

    private void formWindowDeiconified(java.awt.event.WindowEvent evt) {//GEN-FIRST:event_formWindowDeiconified
        // TODO add your handling code here:

		restoreStateOfOpenWindows();

    }//GEN-LAST:event_formWindowDeiconified

    private void jMI_licenseActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_jMI_licenseActionPerformed
        // TODO add your handling code here:

		LicenseJDialog ljd = new LicenseJDialog( this, getAppliConf(), false );
		ljd.setVisible(true);
		boolean closeWindow = true;
		ljd.formWindowClosing(closeWindow);

    }//GEN-LAST:event_jMI_licenseActionPerformed

	/**
	 * @param args the command line arguments
	 */
	public static void main(String args[]) {
		/* Set the Nimbus look and feel */
        //<editor-fold defaultstate="collapsed" desc=" Look and feel setting code (optional) ">
        /* If Nimbus (introduced in Java SE 6) is not available, stay with the default look and feel.
		 * For details see http://download.oracle.com/javase/tutorial/uiswing/lookandfeel/plaf.html 
		 */
/*
		try {
			for (javax.swing.UIManager.LookAndFeelInfo info : javax.swing.UIManager.getInstalledLookAndFeels()) {
				if ("Nimbus".equals(info.getName())) {
					javax.swing.UIManager.setLookAndFeel(info.getClassName());
					break;
				}
			}
		} catch (ClassNotFoundException ex) {
			java.util.logging.Logger.getLogger(MainWindow.class.getName()).log(java.util.logging.Level.SEVERE, null, ex);
		} catch (InstantiationException ex) {
			java.util.logging.Logger.getLogger(MainWindow.class.getName()).log(java.util.logging.Level.SEVERE, null, ex);
		} catch (IllegalAccessException ex) {
			java.util.logging.Logger.getLogger(MainWindow.class.getName()).log(java.util.logging.Level.SEVERE, null, ex);
		} catch (javax.swing.UnsupportedLookAndFeelException ex) {
			java.util.logging.Logger.getLogger(MainWindow.class.getName()).log(java.util.logging.Level.SEVERE, null, ex);
		}
        //</editor-fold>
*/
		
		try
		{
				javax.swing.UIManager.setLookAndFeel( new ToolTipLookAndFeel() );
		} catch (javax.swing.UnsupportedLookAndFeelException ex) {
			java.util.logging.Logger.getLogger(MainWindow.class.getName()).log(java.util.logging.Level.SEVERE, null, ex);
		}
/*
		try
		{
			System.setOut(outputFile("out.txt" ) );
			System.setErr(outputFile("err.txt" ) );
		}
		catch( Throwable th )
		{
			th.printStackTrace();
		}
*/
		/* Create and display the form */
		java.awt.EventQueue.invokeLater(new Runnable() {
			public void run() {
				try
				{
					new MainWindow().setVisible(true);
				}
				catch( Throwable th )
				{
					th.printStackTrace();
				}
			}
		});
	}

   protected static PrintStream outputFile(String name) throws IOException
   {
       return new PrintStream(new BufferedOutputStream(new FileOutputStream(name)), true);
   }
	
    // Variables declaration - do not modify//GEN-BEGIN:variables
    private javax.swing.JCheckBox jCB_whitePlaysFromBottom;
    private javax.swing.JMenuItem jMI_about;
    private javax.swing.JMenuItem jMI_copyFEN;
    private javax.swing.JMenuItem jMI_copyPGN;
    private javax.swing.JMenuItem jMI_editComment;
    private javax.swing.JMenuItem jMI_editInitialPosition;
    private javax.swing.JMenuItem jMI_editTAGsWindow;
    private javax.swing.JMenuItem jMI_exit;
    private javax.swing.JMenuItem jMI_license;
    private javax.swing.JMenuItem jMI_open;
    private javax.swing.JMenuItem jMI_openCurrentGameInDetachedWindow;
    private javax.swing.JMenuItem jMI_paste;
    private javax.swing.JMenuItem jMI_preferences;
    private javax.swing.JMenuItem jMI_save;
    private javax.swing.JMenuItem jMI_saveAs;
    private javax.swing.JMenu jM_Edit;
    private javax.swing.JMenu jM_File;
    private javax.swing.JMenu jM_about;
    private javax.swing.JMenu jM_languageSubmenu;
    private javax.swing.JMenu jM_preferences;
    private javax.swing.JMenu jM_windows;
    private javax.swing.JMenuBar jMenuBar1;
    private javax.swing.JPanel jPanel1;
    private javax.swing.JPanel jPanel3;
    private javax.swing.JPanel jPanel4;
    private javax.swing.JPanel jPanelChessBoardContainer;
    private javax.swing.JPanel jPanelNavigatorContainer1;
    private javax.swing.JRadioButtonMenuItem jRadioButtonMenuItemIdiomaEn;
    private javax.swing.JRadioButtonMenuItem jRadioButtonMenuItemIdiomaEs;
    private javax.swing.JRadioButtonMenuItem jRadioButtonMenuItemIdiomaOtro;
    private javax.swing.JSplitPane jSPmainSplit;
    private javax.swing.JSplitPane jSPnavigatorSplit;
    private javax.swing.JScrollPane jScrollPane1;
    private javax.swing.JScrollPane jScrollPane2;
    private javax.swing.JPopupMenu.Separator jSeparator1;
    private javax.swing.JPopupMenu.Separator jSeparator10;
    private javax.swing.JPopupMenu.Separator jSeparator2;
    private javax.swing.JSeparator jSeparator3;
    private javax.swing.JPopupMenu.Separator jSeparator4;
    private javax.swing.JPopupMenu.Separator jSeparator5;
    private javax.swing.JPopupMenu.Separator jSeparator6;
    private javax.swing.JPopupMenu.Separator jSeparator7;
    private javax.swing.JPopupMenu.Separator jSeparator9;
    // End of variables declaration//GEN-END:variables


	protected void save()
	{
		String fileName = null;
		if( ( _pgnFile == null ) || ( _pgnFile.getFileName() == null ) )
			fileName = showFileChooserDialog( DialogsWrapper.SAVE );
		else
			fileName = _pgnFile.getFileName();

		if( fileName != null )
			saveAs( fileName );
	}
	
	/**
	 * 
	 * @param typeOfDialog		This parameter may take the values: DialogsWrapper.SAVE or DialogsWrapper.OPEN
	 * @return 
	 */
	protected String showFileChooserDialog( int typeOfDialog )
	{
		FilterForFile fnef1 = new FilterForFile( ".pdf : Portable document format", "pdf" );
		FilterForFile fnef2 = new FilterForFile( ".pgn : Chess game notation", "pgn" );

		List<FilterForFile> ffl = new ArrayList<FilterForFile>();
		ffl.add( fnef1 );
		ffl.add( fnef2 );

		String defaultFileName = null;

		if( typeOfDialog == DialogsWrapper.SAVE )
		{
			if( _pgnFile != null )
				defaultFileName = _pgnFile.getFileName();
		}

		String fileName = GenericFunctions.instance().getDialogsWrapper().showFileChooserDialog( this,
																					typeOfDialog,
																					ffl, defaultFileName );

		return( fileName );
	}

	protected PgnChessFile getPgnFile()
	{
		if( _pgnFile == null )
		{
//			_pgnFile = new PgnChessFile( ApplicationConfiguration.instance().getChessLanguageConfiguration() );
			_pgnFile = new PgnChessFile();
		}
		return( _pgnFile );
	}
	
	public Component getChessBoardContainer()
	{
		return( jPanelChessBoardContainer );
	}

	@Override
	public void newPositionInTheMovesTree( List<ChessGameMove> lcgm )
	{
		try
		{
			MoveTreeNode node = _currentChessGame.findNode( lcgm );

			if( node == null )
			{
				_currentChessGame.insertMoves( lcgm );
			}
			_currentChessGame.setCurrentListOfMoves( lcgm );
			_currentChessGame.updateAdditionalInfo();
			_listOfGamesHasBeenModified = true;

			boolean everyThing = true;
			_multiWindowGameManager.updateChessGameWindows(_currentChessGame, everyThing);
//			jTextPane1.update( everyThing );

			EditCommentFrame.instance().setMove( _currentChessGame, _currentChessGame.getCurrentMove() );
		}
		catch( Throwable th )
		{
			th.printStackTrace();
		}
	}

	public void M_releaseResources()
	{
		a_intern=null;	// for the garbage collector to free the memory of the internationallization object and after the memory of this form
	}

	protected String getDescriptionOfGame( ChessGame cg )
	{
		return( cg.getChessGameHeaderInfo().getDescriptionOfGame() );
	}

	protected void changeCurrentChessGame( ChessGame chessGame )
								throws ChessMoveException, ChessPieceCreationException,
										ChessGamePositionException, ChessModelException,
										ChessWriterException
	{
//		if( chessGame != _currentChessGame )
		{
			_currentChessGame = chessGame;

			setTitle(getDescriptionOfGame( _currentChessGame ) );
			
			_currentChessGame.setChessViewConfiguration( ApplicationConfiguration.instance() );
			_currentChessGame.updateAdditionalInfo();

//			_currentChessGame.start();
			
			_chessBoardPanel.setChessBoard( _currentChessGame.getChessBoard() );

			jTextPane1.setChessGame( _currentChessGame );
			boolean everyThing = true;
//			jTextPane1.update( everyThing );
			jTextPane1.updateEverythingFully();

			_multiWindowGameManager.updateChessGameWindowsFully(chessGame);
			
			jTListOfGames.updateTableRow( _currentChessGame );

			if( EditTAGsJFrame.instance() != null )
				EditTAGsJFrame.instance().setChessGame( _currentChessGame );
			
			if( EditCommentFrame.instance() != null )
				EditCommentFrame.instance().setMove( _currentChessGame,
													( _currentChessGame != null ?
														_currentChessGame.getCurrentMove() :
														null )
													);
		}
	}

	public void changeListOfChessGames( List<ChessGame> list, int indexForCurrentGame )
								throws ChessMoveException, ChessPieceCreationException,
										ChessGamePositionException, ChessModelException,
										ChessWriterException
	{
		if( ( list != null ) && ( list.size() > 0 ) )
		{
//			System.out.println( "actualizando lista de partidas ..." );

			_multiWindowGameManager.closeAllWindows();

//			if( _listOfGames != list )
			{
				_listOfGamesHasBeenModified = false;
				_listOfGames = list;
				jTListOfGames.setChessGameList(list);

//			System.out.println( "Actualizando partida ..." );

				indexForCurrentGame = IntegerFunctions.min( list.size()-1, indexForCurrentGame );
				changeCurrentChessGame( _listOfGames.get(indexForCurrentGame) );
			
//			System.out.println( "Partida actualizada ..." );
			}
		}
	}

	public void openFile( String fileName )
	{
		_lastFileName = fileName;
		String extension = FileFunctions.getExtension( fileName ).toLowerCase();

		Thread loadThread = null;
		if( extension.equals( "pgn" ) )
			loadThread = new LoadPGNThread( this, fileName );
		else if( extension.equals( "pdf" ) )
		{
			loadThread = new LoadPDFThread( this, fileName );
//			jCB_renderSegments.setSelected( false );
		}

		loadThread.start();
	}

	public void saveAs( String fileName )
	{
		String extension = "pgn";
		if( fileName != null )
		{
			_lastFileName = fileName;
			extension = FileFunctions.getExtension( fileName ).toLowerCase();
		}

		if( (_listOfGames == null ) || ( _listOfGames.size() == 0 ) )
		{
			_listOfGames = new ArrayList< ChessGame >();
			_listOfGames.add( jTextPane1.getChessGame() );
		}

		Thread saveThread = null;
		if( extension.equals( "pgn" ) )
			saveThread = new SavePGNThread( this, getPgnFile(), _listOfGames, fileName );

		saveThread.run();
	}

	@Override
	public void newChessGameChosen( ChessGame chessGame, boolean hasBeenModified )
	{
		try
		{
			_listOfGamesHasBeenModified = hasBeenModified;
			changeCurrentChessGame( chessGame );
		}
		catch( Throwable th )
		{
			th.printStackTrace();
			GenericFunctions.instance().getDialogsWrapper().showMessageDialog(this, th.getMessage(),
											getAppStrConf().getProperty(AppStringsConf.CONF_PROBLEM_SWITCHING_TO_NEW_GAME ),
											DialogsWrapper.ERROR_MESSAGE );
		}
	}

	@Override
	public void editInitialPosition( ChessGame cg )
	{
		ChessGamePosition oldPosition = cg.getInitialPosition();
		int oldLevel = cg.getMoveTreeGame().getLevel();

		InitialPositionDialog.instance().setChessGame( cg );

		InitialPositionDialog.instance().setVisible( true );

		ChessGamePosition newPosition = InitialPositionDialog.instance().getPositionResult();
		if( newPosition != null )
		{
			int moveNumber = 1;
			if( cg.getMoveTreeGame().getNumberOfChildren() > 0 )
				moveNumber = cg.getMoveTreeGame().getChild(0).getMoveNumber();

//			newPosition.setMoveNumberToInitialPosition( moveNumber );

			cg.setInitialPosition(newPosition);
			try
			{
				cg.start();
				cg.updateAdditionalInfo();
				newChessGameChosen( cg, true );
			}
			catch( Throwable th )
			{
				th.printStackTrace();
				cg.setInitialPosition(oldPosition);
				try
				{
					cg.updateAdditionalInfo();
				}
				catch( Throwable th2 )
				{
					th2.printStackTrace();
				}
				if( oldPosition == null )
					cg.getMoveTreeGame().setLevel( oldLevel );
			}
		}
	}

	@Override
	public void editComment( ChessGame cg, MoveTreeNode mtn )
	{
		EditCommentFrame.instance().setMove( cg, mtn );
		EditCommentFrame.instance().setVisible(true);
		EditCommentFrame.instance().setState ( Frame.NORMAL );
	}

	protected void parsePastedText( String text )
	{
		boolean success = false;
		List<ChessGame> listOfNewGames = null;
		ChessGame newGame = null;

		try
		{
			PGNChessGameParser pgnParser = new PGNChessGameParser( EnglishChessConfiguration.instance() );

			listOfNewGames = pgnParser.parseChessGameText( text, null );
			success = true;
		}
		catch( Throwable th )
		{
		}

		if( !success )
		{
			try
			{
				ChessGamePosition cgp = new ChessGamePosition( text );
				newGame = new ChessGame( getAppliConf() );
				
				newGame.setInitialPosition( cgp );

				success = true;
			}
			catch( Throwable th )
			{
			}
		}

		if( !success )
		{
			try
			{
				RawChessGameParser rcgp = new RawChessGameParser( getAppliConf().getChessLanguageConfigurationToParseTextFrom(),
																	null );

				listOfNewGames = rcgp.parseChessGameText( text, null );

				success = true;
			}
			catch( Throwable th )
			{
			}
		}

		if( !success )
		{
			showError( getAppStrConf().getProperty( AppStringsConf.CONF_ERROR ),
					getAppStrConf().getProperty( AppStringsConf.CONF_ERROR_SCANNING_TEXT ) );			
		}
		else
		{
			int indexForCurrentGame = _listOfGames.size();
			
			if( listOfNewGames != null )
			{
				int index = 1;
				Iterator<ChessGame> it = listOfNewGames.iterator();
				while( it.hasNext() )
				{
					it.next().getChessGameHeaderInfo().put( ChessGameHeaderInfo.EVENT_TAG, "Pasted game - " + index );
					index++;
				}

				_listOfGames.addAll( listOfNewGames );
			}

			if( newGame != null )
			{
				newGame.getChessGameHeaderInfo().put( ChessGameHeaderInfo.EVENT_TAG, "Position Pasted" );
				_listOfGames.add( newGame );
			}

			try
			{
				changeListOfChessGames( _listOfGames, indexForCurrentGame );
				_listOfGamesHasBeenModified = true;	// this line must be the last one, because in changeListOfChessGames, it is restores to false.
			}
			catch( Throwable th )
			{
				th.printStackTrace();
			}
		}
	}

	protected void parsePastedText()
	{
		String pasteText = SystemClipboard.instance().getClipboardContents();
		if( pasteText != null )
			parsePastedText( pasteText );
		else
			GenericFunctions.instance().getDialogsWrapper().showMessageDialog( this,
											getAppStrConf().getProperty( AppStringsConf.CONF_CLIPBOARD_CONTENT_NOT_A_STRING ),
											getAppStrConf().getProperty( AppStringsConf.CONF_ERROR_PASTING_TEXT ),
												DialogsWrapper.ERROR_MESSAGE );
	}

	@Override
	public void block()
	{
		jM_File.setEnabled( false );
		jM_Edit.setEnabled( false );
		jM_preferences.setEnabled( false );
		jM_windows.setEnabled( false );
		jM_about.setEnabled( false );

		changeToWaitCursor();
		_multiWindowGameManager.changeToWaitCursor();
		EditTAGsJFrame.instance().changeToWaitCursor();
		EditCommentFrame.instance().changeToWaitCursor();

		if( _pdfViewerWindow != null )
			_pdfViewerWindow.block();
	}

	@Override
	public void unblock()
	{
		jM_File.setEnabled( true );
		jM_Edit.setEnabled( true );
		jM_preferences.setEnabled( true );
		jM_windows.setEnabled( true );
		jM_about.setEnabled( true );

		revertChangeToWaitCursor();
		_multiWindowGameManager.revertChangeToWaitCursor();
		EditTAGsJFrame.instance().revertChangeToWaitCursor();
		EditCommentFrame.instance().revertChangeToWaitCursor();

		if( _pdfViewerWindow != null )
			_pdfViewerWindow.unblock();
	}

	@Override
	public void showError( String message, String title )
	{
		GenericFunctions.instance().getDialogsWrapper().showMessageDialog( this, message, title,
															DialogsWrapper.ERROR_MESSAGE );
	}
	
	@Override
	public void newChessListGameLoaded( List<ChessGame> list, PgnChessFile pgnFile )
	{
		try
		{
//			System.out.println( "newChessListGameLoaded ..." );

			changeListOfChessGames( list, 0 );
			_pgnFile = pgnFile;
		}
		catch( Throwable th )
		{
			th.printStackTrace();
			showError( th.getMessage(), "Error while loading the new list of games" );
		}
	}


	@Override
	public void navigator_start( InformerInterface panel )
	{
        try
        {
            _currentChessGame.start();
			boolean everyThing = false;
            jTextPane1.update( everyThing );
        }
        catch( Throwable th )
        {
            th.printStackTrace();
        }
	}

	@Override
	public void navigator_end( InformerInterface panel )
	{
        try
        {
            _currentChessGame.end();
			boolean everyThing = false;
            jTextPane1.update( everyThing );
        }
        catch( Throwable th )
        {
            th.printStackTrace();
        }
	}

	@Override
	public void navigator_previous( InformerInterface panel )
	{
        try
        {
            _currentChessGame.back();
			boolean everyThing = false;
            jTextPane1.update( everyThing );
        }
        catch( Throwable th )
        {
            th.printStackTrace();
        }
	}

	@Override
	public void navigator_next( InformerInterface panel )
	{
        try
        {
            _currentChessGame.forward();
			boolean everyThing = false;
            jTextPane1.update( everyThing );
        }
        catch( Throwable th )
        {
            th.printStackTrace();
        }
	}

	@Override
	public void newChessPDFLoaded(PDDocument pdfDocument)
	{
		if( _pdfViewerWindow == null )
			_pdfViewerWindow = new PDFviewerWindow(this, this);

		_pdfViewerWindow.setNewPDF( _lastFileName, pdfDocument );
		
		try
		{
			_pdfViewerWindow.changeLanguage( getAppliConf().getLanguage() );
		}
		catch( Throwable th )
		{
			th.printStackTrace();
		}

		_pdfViewerWindow.setVisible( true );
		if( _pdfViewerWindow.getState() == Frame.ICONIFIED )
			_pdfViewerWindow.setState( Frame.NORMAL );
	}

	@Override
	public void cancel()
	{
	}

	/** Handle the key-released event from the text field. */
	@Override
    public void keyReleased(KeyEvent e)
	{
    }

	/** Handle the key typed event from the text field. */
	@Override
    public void keyTyped(KeyEvent event)
	{
    }

    /** Handle the key-pressed event from the text field. */
	@Override
    public void keyPressed(KeyEvent event)
	{
		if( (event.getKeyCode() == KeyEvent.VK_UP)
			|| (event.getKeyCode() == KeyEvent.VK_KP_UP) )
		{
			goToPreviousVariation();
		}
		else if (event.getKeyCode() == KeyEvent.VK_DOWN
			|| (event.getKeyCode() == KeyEvent.VK_KP_DOWN) )
		{
			goToNextVariation();
		}
		else if (event.getKeyCode() == KeyEvent.VK_LEFT
			|| (event.getKeyCode() == KeyEvent.VK_KP_LEFT) )
		{
			goToPreviousMove();
		}
		else if (event.getKeyCode() == KeyEvent.VK_RIGHT
			|| (event.getKeyCode() == KeyEvent.VK_KP_RIGHT) )
		{
			goToNextMove();
		}
    }

	protected void setNewCurrentMoveTreeNode( MoveTreeNode mtn )
	{
		if( mtn != null )
		{
			try
			{
				_currentChessGame.setCurrentListOfMoves( mtn.getGameMoveList() );
				EditCommentFrame.instance().setMove( _currentChessGame, _currentChessGame.getCurrentMove() );

				boolean everyThing = false;
	            jTextPane1.update( everyThing );
			}
			catch( Throwable th )
			{
				th.printStackTrace();
				GenericFunctions.instance().getDialogsWrapper().showMessageDialog( this,
										getAppStrConf().getProperty( AppStringsConf.CONF_ERROR_CHANGING_CURRENT_MOVE ) + th.getMessage(),
										getAppStrConf().getProperty( AppStringsConf.CONF_ERROR_BROWSING_THE_TREE ),
										DialogsWrapper.ERROR_MESSAGE );
			}
		}
	}

	protected void goToPreviousVariation()
	{
		if( _currentChessGame != null )
		{
			MoveTreeNode newPositionToBeSet = _currentChessGame.getCurrentMove().getPreviousVariationMoveNode();
			setNewCurrentMoveTreeNode(newPositionToBeSet);
		}
	}

	protected void goToNextVariation()
	{
		if( _currentChessGame != null )
		{
			MoveTreeNode newPositionToBeSet = _currentChessGame.getCurrentMove().getNextVariationMoveNode();
			setNewCurrentMoveTreeNode(newPositionToBeSet);
		}
	}

	protected void goToPreviousMove()
	{
		if( _currentChessGame != null )
		{
			MoveTreeNode newPositionToBeSet = _currentChessGame.getCurrentMove().getParent();
			setNewCurrentMoveTreeNode(newPositionToBeSet);
		}
	}

	protected void goToNextMove()
	{
		if( _currentChessGame != null )
		{
			MoveTreeNode newPositionToBeSet = _currentChessGame.getCurrentMove().getChild(0);
			setNewCurrentMoveTreeNode(newPositionToBeSet);
		}
	}

	protected void setKeyListener( Component comp )
	{
		if( comp instanceof Container	)
		{
			Container contnr = (Container) comp;
			if( !( contnr instanceof JMenu ) )
			{
				for( int ii=0; ii<contnr.getComponentCount(); ii++ )
				{
					setKeyListener( contnr.getComponent(ii) );
				}
			}
		}

		comp.addKeyListener(this);
	}
/*
	protected void marcaIdiomaEnMenu( String idioma )
	{
		try
		{
			if( ( idioma.compareTo(jRadioButtonMenuItemIdiomaEs.getText()) == 0 ) &&
				! jRadioButtonMenuItemIdiomaEs.isSelected() )
					jRadioButtonMenuItemIdiomaEs.setSelected(true);
			else if ( ( idioma.compareTo( jRadioButtonMenuItemIdiomaEn.getText()) == 0 )  &&
				! jRadioButtonMenuItemIdiomaEn.isSelected() )
					jRadioButtonMenuItemIdiomaEn.setSelected(true);
			else if ( ( idioma.compareTo( jRadioButtonMenuItemIdiomaOtro.getText() ) == 0 )  &&
				! jRadioButtonMenuItemIdiomaOtro.isSelected() )
					jRadioButtonMenuItemIdiomaOtro.setSelected(true);
		}
		catch( Exception ex )
		{
			ex.printStackTrace();
		}
	}
*/
	protected void applyNewApplicationConfiguration()
	{
		setLanguageInMenu( getAppliConf().getLanguage() );

		_chessBoardPanel.refresh();
		_multiWindowGameManager.applyNewConfiguration();

//		float factor = getAppliConf().M_getApplicationFontSize();
//		M_changeFontSizeInApplication( factor );
	}
/*
	protected void M_changeFontSizeInApplication( float factor )
	{
		if( a_intern != null )
		{
			a_intern.M_changeFontSize( factor );
		}

		JInternalFrame[] ifArray = jMainDesktopPane.getAllFrames();
		for( int ii=0; ii<ifArray.length; ii++ )
		{
			if( ifArray[ii] instanceof FileJInternalFrame)
			{
				FileJInternalFrame jif = (FileJInternalFrame) ifArray[ii];
				
				jif.M_changeFontSize( factor );
			}
		}
	}
*/

	public void formWindowClosing( boolean closeWindow )
	{
		_multiWindowGameManager.closeAllWindows();
		if( _pdfViewerWindow != null )
			_pdfViewerWindow.closeWindow();

		if( closeWindow )
		{
			EditCommentFrame.instance().formWindowClosing(closeWindow);
			EditTAGsJFrame.instance().formWindowClosing(closeWindow);
			InitialPositionDialog.instance().formWindowClosing(closeWindow);
		}

//		setConfigurationChanges();

		if( _appliConf != null )
		{
			try
			{
				_appliConf.M_saveConfiguration();
			}
			catch( Throwable th )
			{
				th.printStackTrace();
			}
		}

		if( getAppStrConf() != null )
		{
			try
			{
				getAppStrConf().M_saveConfiguration();
			}
			catch( Throwable th )
			{
				th.printStackTrace();
			}
		}

		if( ChessStringsConf.instance() != null )
		{
			try
			{
				ChessStringsConf.instance().M_saveConfiguration();
			}
			catch( Throwable th )
			{
				th.printStackTrace();
			}
		}

		super.formWindowClosing( closeWindow );

		if( closeWindow )
			System.exit(0);
	}

/*
	// https://blogs.oracle.com/moonocean/entry/change_properties_of_java_tooltip
	static class ToolTipLookAndFeel extends MetalLookAndFeel
	{
		protected void initSystemColorDefaults(UIDefaults table)
		{        
			super.initSystemColorDefaults(table);        
			table.put("info", new ColorUIResource(255, 247, 200));    
		}

		protected void initComponentDefaults(UIDefaults table) {
			super.initComponentDefaults(table);

		Border border = BorderFactory.createLineBorder(new Color(76,79,83));
		table.put("ToolTip.border", border);
		}
	}
*/
/*
	protected void setConfigurationChanges() throws ConfigurationException
	{
		getAppliConf().changeLanguage(_language );
	}
*/
	protected AppStringsConf getAppStrConf()
	{
		return( _appStrConf );
	}

	protected void setWindowsNotAlwaysOnTop()
	{
		_multiWindowGameManager.setWindowsAlwaysOnTop( false );
		EditCommentFrame.instance().setAlwaysOnTop( false );
		EditTAGsJFrame.instance().setAlwaysOnTop( false );
	}

	protected void restoreWindowsAlwaysOnTop()
	{
		_multiWindowGameManager.setWindowsAlwaysOnTop( getAppliConf().getDetachedGameWindowsAlwaysOnTop() );
		EditCommentFrame.instance().updateAlwaysOnTop();
		EditTAGsJFrame.instance().updateAlwaysOnTop();
	}

	public void setHasBeenModified( boolean value )
	{
		_listOfGamesHasBeenModified |= value;
	}

	public boolean checkToSaveCurrentPGN()
	{
		boolean result = true;

		if( _listOfGamesHasBeenModified )
		{
			Object[] options = new Object[] {
				getInternationalString( CONF_YES ),
				getInternationalString( CONF_NO ),
				getInternationalString( CONF_CANCEL )
			};

			int answer = GenericFunctions.instance().getDialogsWrapper().showOptionDialog( this,
														getInternationalString( CONF_CURRENT_LIST_OF_GAMES_HAS_BEEN_MODIFIED_DO_YOU_WANT_TO_SAVE_CHANGES ),
														getInternationalString( CONF_WARNING ),
														DialogsWrapper.YES_NO_CANCEL_OPTION,
														DialogsWrapper.WARNING_MESSAGE,
														options,
														null );

			if( ( answer == -1 ) || ( answer == 2 ) )		// cancel
			{
				result = false;
			}
			else if( answer == 0 )		// yes
			{
				save();
				result = !_listOfGamesHasBeenModified;
			}
			else		// no
				_listOfGamesHasBeenModified = false;		// we mark the list of games as not modified.
		}

		return( result );
	}

	protected void saveStateOfWindowAndSetNewState( Frame frame, Integer state )
	{
		if( frame != null )
		{
			int previousState = frame.getState();

			if( previousState != Frame.ICONIFIED )
				_stateOfWindows.put( frame, previousState );

			if( state != null )
				frame.setState( state );
		}
	}

	protected void saveStateOfOpenWindowsAndIconifyThem()
	{
		_stateOfWindows = new HashMap< Frame, Integer >();

		saveStateOfWindowAndSetNewState( EditCommentFrame.instance(), Frame.ICONIFIED );
		saveStateOfWindowAndSetNewState( EditTAGsJFrame.instance(), Frame.ICONIFIED );
		saveStateOfWindowAndSetNewState( _pdfViewerWindow, Frame.ICONIFIED );
		
		_multiWindowGameManager.saveStateOfOpenWindowsAndIconifyThem();
	}

	protected void restoreStateOfOpenWindows()
	{
		if( _stateOfWindows != null )
		{
			Iterator< Map.Entry< Frame, Integer> > it = _stateOfWindows.entrySet().iterator();

			while( it.hasNext() )
			{
				Map.Entry< Frame, Integer > entry = it.next();
				Frame frame = entry.getKey();

				if( frame.isVisible() )
					frame.setState( entry.getValue() );
			}

			_stateOfWindows = null;

			_multiWindowGameManager.restoreStateOfOpenWindows();
		}
	}
}
