/* 
 * Copyright (C) 2017 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */
package com.frojasg1.chesspdfbrowser.io.parsers.implementation;

import com.frojasg1.chesspdfbrowser.configuration.ChessStringsConf;
import com.frojasg1.chesspdfbrowser.exception.ChessParserException;
import com.frojasg1.chesspdfbrowser.io.parsers.tokens.MoveToken;
import com.frojasg1.chesspdfbrowser.io.parsers.tokens.PreToken;
import com.frojasg1.chesspdfbrowser.io.parsers.tokens.Token;
import com.frojasg1.chesspdfbrowser.io.parsers.tokens.TokenId;
import com.frojasg1.chesspdfbrowser.model.chess.configuration.ChessLanguageConfiguration;
import com.frojasg1.chesspdfbrowser.model.chess.game.tree.NAG;
import com.frojasg1.chesspdfbrowser.pdf.textsegmentation.PDFPageSegmentatorInterface;
import com.frojasg1.general.BooleanReference;
import com.frojasg1.general.number.IntegerReference;
import com.frojasg1.general.progress.CancellationException;
import com.frojasg1.general.progress.UpdatingProgress;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.regex.Matcher;

/**
 *
 * @author Usuario
 */
public class RawLexicalAnalyser extends PGNLexicalAnalyser
{
	protected static final int MINIMUM_NUMBER_OF_TIMES_A_TITLE_MUST_BE_REPEATED = 7;
	
	protected boolean _isLastString = false;
	protected UpdatingProgress _updatingProgress = null;

	protected int[] _positionsOfPageStarts = null;

	protected PDFPageSegmentatorInterface _source = null;

	protected int _currentPageNumber = -1;

	protected String _lineStringToSkip = null;
	
	public RawLexicalAnalyser( ChessLanguageConfiguration clc, PDFPageSegmentatorInterface source, UpdatingProgress up )
		throws ChessParserException, CancellationException
	{
		super( clc );

		_source = source;
		_updatingProgress = up;

		setStringToParse( getTextToParse( source ) );

		setHasToBeStrict( false );
		setHasToNotifyReturns( false );
		setHasToNotifyBlankLines( false );
	}

	public RawLexicalAnalyser( ChessLanguageConfiguration clc, String text )
		throws ChessParserException, CancellationException
	{
		super( clc );

		setStringToParse( getTextToParse( text ) );

		setHasToBeStrict( false );
		setHasToNotifyReturns( false );
		setHasToNotifyBlankLines( false );
	}

	public int getPageOfPosition( int position )
	{
		int ini = 0;
		int end = _positionsOfPageStarts.length - 1;
		int result = -1;
		while( result == -1 )
		{
			int mid = ( ini + end ) / 2;
			
			if( matchesPageFromPosition( ini, position ) )
			{
				result = ini;
			}
			else if( matchesPageFromPosition( end, position ) )
			{
				result = end;
			}
			else if( matchesPageFromPosition( mid, position ) )
			{
				result = mid;
			}
			else if( ( end - ini ) < 2 )
			{
				result = mid;
				System.out.println( "Error, page not found. Position: " + position + ". We break the loop to avoid infinite looping" );
			}
			else if( _positionsOfPageStarts[mid] >= position )
			{
				end = mid;
			}
			else
			{
				ini = mid;
			}
		}
		return( result );
	}

	protected String nextLine()
	{
		String result = null;

		while( result == null )
		{
			result = super.nextLine();
			if( result == null )
				break;

			if( ( _currentPageNumber < ( _positionsOfPageStarts.length - 1 ) ) &&
				(_newLineMatcher.start() == _positionsOfPageStarts[_currentPageNumber + 1] ) )
			{
				_currentPageNumber++;
				if( ( _lineStringToSkip != null ) &&
					( _lineStringToSkip.equals( result ) ) )
				{
					result = null;
				}
			}
		}

		return( result );
	}

	protected boolean matchesPageFromPosition( int arrayIndex, int position )
	{
		boolean result = ( _positionsOfPageStarts[ arrayIndex ] <= position );
		if( arrayIndex < (_positionsOfPageStarts.length-1) )
			result = result && ( _positionsOfPageStarts[arrayIndex+1] > position );
		
		return( result );
	}

	protected String getTextToParse(String text) throws ChessParserException, CancellationException
	{
		List<Integer> beginningOfPages = new ArrayList<Integer>();

		_positionsOfPageStarts = new int[1];
		_positionsOfPageStarts[0] = text.length();
		
		return( text );
	}

	protected void addStringToCandidateToSkip( Map<String, IntegerReference> map, String pageString )
	{
		Matcher newLineMatcher = _newLinePattern.matcher( pageString );
		String line = null;
		if( newLineMatcher.find() )
		{
			if( newLineMatcher.groupCount() == 2 )
			{
				line = newLineMatcher.group( 1 );
			}
		}
		
		if( line != null )
		{
			IntegerReference value = map.get( line );
			if( value == null )
				map.put( line, new IntegerReference( 1 ) );
			else
			{
				value._value++;
			}
		}
	}

	protected String getStringToSkip( Map<String, IntegerReference> hashCandidateToSkip )
	{
		String result = null;
		int max = 0;
		
		Iterator<Map.Entry<String, IntegerReference> > it = hashCandidateToSkip.entrySet().iterator();
		while( it.hasNext() )
		{
			Map.Entry<String, IntegerReference> mapElem = it.next();

			if( ( mapElem.getValue()._value > max ) &&
				( mapElem.getValue()._value >= MINIMUM_NUMBER_OF_TIMES_A_TITLE_MUST_BE_REPEATED )
			  )
			{
				max = mapElem.getValue()._value;
				result = mapElem.getKey();
			}
		}

		return( result );
	}

	protected String getTextToParse(PDFPageSegmentatorInterface source) throws ChessParserException, CancellationException
	{
		StringBuilder result = new StringBuilder();

		// this map counts the times a string appear as the first line of the page.
		// we want to skip that line if it is repeated many times (then it is a title, and it is not part of the games)
		Map<String, IntegerReference> hashCandidateToSkip = new HashMap<String, IntegerReference>();

		List<Integer> beginningOfPages = new ArrayList<Integer>();

		if( _updatingProgress != null )
			_updatingProgress.up_childStarts();

		beginningOfPages.add(0);
		for( int ii=0; ii<_source.getNumberOfPages(); ii++ )
		{
			try
			{
				if( _updatingProgress != null )
					_updatingProgress.up_updateProgressFromChild( (double) ii / _source.getNumberOfPages() );

				if( ii>0 )
				{
					result.append( "\n" );
					beginningOfPages.add( result.length() );
				}

				String pageString = source.getTextOfPage( ii );
				result.append( pageString );
				addStringToCandidateToSkip( hashCandidateToSkip, pageString );
			}
			catch( CancellationException ce )
			{
				throw( ce );
			}
			catch( Throwable th )
			{
				th.printStackTrace();
				throw( new ChessParserException( th.toString() ) );
			}
		}

		beginningOfPages.add( result.length() );

		if( _updatingProgress != null )
			_updatingProgress.up_childEnds();

		_positionsOfPageStarts = new int[beginningOfPages.size()];
		Iterator<Integer> it = beginningOfPages.iterator();
		int index = 0;
		while( it.hasNext() )
		{
			_positionsOfPageStarts[index] = it.next();
			index++;
		}
		
		_lineStringToSkip = getStringToSkip( hashCandidateToSkip );
		
		return( result.toString() );
	}

	@Override
	protected boolean isEndOfPreToken( char currentChar )
	{
		boolean result = isBlank( currentChar );

		if( ! result )
		{
			switch( currentChar )
			{
				case ' ':
				case '{':
				case '}':
				case '.':
				case '(':
				case ')':
				case '[':
				case ']':
				case '\t':
				case ';':
				case ':':
				case ',':
					result = true;
			}
		}

		return( result );
	}

	@Override
	protected boolean isANag( String text )
	{
		return( NAG.getNAGfromStringToShow( text ) != -1 );
	}

	@Override
	protected TokenId getTokenId_simple( String preTokenStr )
	{
		TokenId result = null;

		if( preTokenStr.equals( "." ) )			result = TokenId.DOT;
//		else if( preTokenStr.equals( "(" ) )	result = TokenId.OPEN_BRACKET;
//		else if( preTokenStr.equals( ")" ) )	result = TokenId.CLOSE_BRACKET;
//		else if( preTokenStr.equals( "{" ) )	result = TokenId.OPEN_BRACE;
//		else if( preTokenStr.equals( "}" ) )	result = TokenId.CLOSE_BRACE;
		else if( isANumber( preTokenStr ) )		result = TokenId.NUMBER;
		else if( isAMove( preTokenStr ) )		result = TokenId.MOVE;
		else if( isAResult( preTokenStr ) )		result = TokenId.RESULT;
		else if( isANag( preTokenStr ) )		result = TokenId.NAG;

		return( result );
	}

	@Override
	protected boolean getNewLine() throws ChessParserException
	{
		while( (_currentLine == null) && !_isEOF )
		{
			_currentLine = nextLine();

			_isEOF = ( _currentLine == null );
		}

		if( _currentLine != null )
		{
			_currentPosInLine = 0;
		}

		return( _currentLine != null );
	}

	/**
	 * We override this method to allow joining the two parts of a string splitted by a new line
	 * As for example can happen in case of a O-O which can be splitted in two parts 
	 * ( O- and O, each one in a different line).
	 * @throws ChessParserException 
	 */
	@Override
	protected void getNewPreToken() throws ChessParserException
	{
		PreToken pt = null;
		PreToken stringWrappedToEndOfLinePt = null;
		PreToken endOfLinePt = null;

		while( !_isEOF && (pt==null) )
		{
			if( _currentLine == null )
			{
				getNewLine();
			}

			if( !_isEOF )
			{
				pt = getNextPreTokenInCurrentLine();

				if( pt != null )
				{
					if( isStringWrappedToEndOfLinePt( pt ) )
					{
						stringWrappedToEndOfLinePt = pt;
						pt = null;
					}
					else if( stringWrappedToEndOfLinePt!=null )
					{
						if( isStringWrappedToStartOfLinePt( pt ) )
						{
							String ptStr = stringWrappedToEndOfLinePt.getString() + pt.getString();
							PreToken tmpPt = getPreTokenFromPreTokenString( ptStr );
							_currentPosInLine--;	//because it had been incremented inside the previous function.

							if( ( tmpPt != null ) && ( tmpPt.getTokenId().equals( TokenId.MOVE ) ) )
							{
								stringWrappedToEndOfLinePt = null;
								endOfLinePt = null;
								pt = tmpPt;
							}
						}
					}
				}

				if( pt != null )
				{
					if( stringWrappedToEndOfLinePt != null )
					{
						_bufferOfInputPreTokens.add( stringWrappedToEndOfLinePt );
						stringWrappedToEndOfLinePt = null;
					}

					if( endOfLinePt != null )
					{
						_bufferOfInputPreTokens.add( endOfLinePt );
						endOfLinePt = null;
					}

					_bufferOfInputPreTokens.add( pt );
				}

				if( _currentPosInLine >= _currentLine.length() )
				{
					_currentLine = null;
					if( hasToNotifyReturns() )
					{
						endOfLinePt = new PreToken( TokenId.RETURN, "\r", _currentLineNumber, _currentPosInLine );
						if( pt != null )
							_bufferOfInputPreTokens.add( endOfLinePt );
					}
				}
			}
		}
	}

	protected boolean isStringWrappedToEndOfLinePt( PreToken pt )
	{
		boolean result = ( pt.getTokenId().equals( TokenId.STRING ) ||
							pt.getString().equals("O-O" ) ||
								pt.getString().equals("0-0" ) ) &&
							(	(pt.getInitialPosition() + pt.getString().length() ) ==
								( _newLineMatcher != null ? _newLineMatcher.start()+_currentLine.length() : 0 ) );

		return( result );
	}

	protected boolean isStringWrappedToStartOfLinePt( PreToken pt )
	{
		boolean result = ( pt.getTokenId().equals( TokenId.STRING ) ||
							pt.getTokenId().equals( TokenId.NUMBER ) ||
							( pt.getString().length()>=3 ) &&
							( pt.getString().substring(0,3).equals("O-O" ) ||
								pt.getString().substring(0,3).equals("0-0" ) ) ) &&
							( pt.getInitialPosition() == ( _newLineMatcher != null ? _newLineMatcher.start() : 0 ) );

		return( result );
	}

	protected void processBufferOfInputPreTokens( boolean hasToEmptyBuffer ) throws ChessParserException
	{
		boolean hasToProcess = ( _bufferOfInputPreTokens.size() > 0 );

		BooleanReference mustLoop = new BooleanReference( false );
		while( hasToProcess )
		{
			Token token = null;

			mustLoop._value = true;
			do
			{
				PreToken preToken = _bufferOfInputPreTokens.getFirst();
				if( ( preToken.getTokenId().equals( TokenId.NUMBER ) ) ||
					( preToken.getTokenId().equals( TokenId.MOVE ) ) )
				{
					token = createTokensForMove( preToken, hasToEmptyBuffer, mustLoop );	// this function must not return token != null and mustLoop==true
				}

				if( mustLoop._value )	// if mustLoop, we must discard the preToken.
				{
					_bufferOfInputPreTokens.removeFirst();
					mustLoop._value = ( _bufferOfInputPreTokens.size() > 0 );
				}
			}
			while( mustLoop._value ); // if we must not loop is because no token has been found, or there are not more preTokens in the bufferOfInputPreTokens.

			hasToProcess = ( ( token != null ) && ( _bufferOfInputPreTokens.size() > 0 ) );
		}
	}

	/**
	 * 
	 * @param preToken
	 * @param hasToEmptyBuffer
	 * @param mustDiscard. is a reference to boolean. At the exit of the function, it must contain if the numberPreToken must be discarded or not
	 * @return
	 * @throws ChessParserException 
	 */
	protected Token createTokensForMove( PreToken preToken, boolean hasToEmptyBuffer, BooleanReference mustDiscard ) throws ChessParserException
	{
		Token result = null;

		int moveNumber = -1;
		
		if( preToken.getTokenId().equals( TokenId.NUMBER) )
			moveNumber = Integer.parseInt( preToken.getString() );

		boolean[] optional4 = { false, true, false, true, false, true, false, true };
		boolean[] optional1 = { false, true, false, true, false, true };
		boolean[] optional3 = { false, false, false, false, true, false, true };
		TokenId[] case4 = { TokenId.MOVE, TokenId.NAG, TokenId.NUMBER, TokenId.DOT, TokenId.MOVE, TokenId.NAG, TokenId.MOVE, TokenId.NAG };
		TokenId[] case5 = { TokenId.MOVE, TokenId.NAG, TokenId.NUMBER, TokenId.DOT, TokenId.MOVE, TokenId.NAG };
		TokenId[] case1 = { TokenId.NUMBER, TokenId.DOT, TokenId.MOVE, TokenId.NAG, TokenId.MOVE, TokenId.NAG };
		TokenId[] case2 = { TokenId.NUMBER, TokenId.DOT, TokenId.MOVE, TokenId.NAG };
		TokenId[] case3 = { TokenId.NUMBER, TokenId.DOT, TokenId.DOT, TokenId.DOT, TokenId.DOT, TokenId.MOVE, TokenId.NAG };

		ResultOfCheckMoveTokenSequence resultTS = new ResultOfCheckMoveTokenSequence();

		if( matchTokenIdsForMove( case4, optional4, hasToEmptyBuffer, resultTS ).getResultListOfPreTokens() != null )
		{
			result = createCase4Move( resultTS, hasToEmptyBuffer );
		}
		else if( (!resultTS.getHasNotFailed() || hasToEmptyBuffer ) &&
					( matchTokenIdsForMove( case5, optional4, hasToEmptyBuffer, resultTS ).getResultListOfPreTokens() != null ) )
		{
			result = createCase5Move( resultTS, hasToEmptyBuffer );
		}
		else if( matchTokenIdsForMove( case1, optional1, hasToEmptyBuffer, resultTS ).getResultListOfPreTokens() != null )
		{
			result = createCase1Move( moveNumber, resultTS, hasToEmptyBuffer );
		}
		else if( (!resultTS.getHasNotFailed() || hasToEmptyBuffer ) &&
					( matchTokenIdsForMove( case2, optional1, hasToEmptyBuffer, resultTS ).getResultListOfPreTokens() != null ) )
		{
			result = createCase2Move( moveNumber, resultTS, hasToEmptyBuffer );
		}
		else if( (!resultTS.getHasNotFailed() || hasToEmptyBuffer ) &&
					( matchTokenIdsForMove( case3, optional3, hasToEmptyBuffer, resultTS ).getResultListOfPreTokens() != null ) )
		{
			result = createCase3Move( moveNumber, resultTS, hasToEmptyBuffer );
		}
		else if( hasToBeStrict() && !resultTS.getHasNotFailed() )
		{
			throw( new ChessParserException( String.format( getChessStrConf().getProperty( ChessStringsConf.CONF_CLOSING_OF_COMMNENT_NOT_FOUND ),
												preToken.getLineNumber(),
												preToken.getString()  ) ) );
		}

		mustDiscard._value = !resultTS.getHasNotFailed();

		if( result != null )
		{
			for( int ii=0; ii<resultTS.getNumberOfInputItemsToErase()._value; ii++ )
				_bufferOfInputPreTokens.removeFirst();
		}

		return( result );
	}

	@Override
	protected MoveToken getMoveTokenFromList( IntegerReference index, LinkedList<PreToken> listOfPreTokens,
												int moveNumber, boolean whiteToPlay, boolean hasToEmptyBuffer )  throws ChessParserException
	{
		PreToken movePreToken = getPreTokenAndIncrementIndex( listOfPreTokens, index );
		if( ( movePreToken != null ) && ( movePreToken.getTokenId() != TokenId.MOVE ) )
		{
			movePreToken = getPreTokenAndIncrementIndex( listOfPreTokens, index );
		}
		PreToken nextPreToken = getPreTokenAndIncrementIndex( listOfPreTokens, index );

		MoveToken result = null;
		if( ( movePreToken != null ) && ( hasToEmptyBuffer || ( nextPreToken != null ) ) )
		{
			result = new MoveToken( movePreToken );
			result.setString( _chessLanguageConfiguration.translateMoveStringToEnglish( result.getString() ) );

			while( ( nextPreToken!= null ) && nextPreToken.getTokenId().equals( TokenId.NAG ) )
			{
				NAG nag = new NAG( nextPreToken.getString() );
				result.addNag(nag);
				nextPreToken = getPreTokenAndIncrementIndex( listOfPreTokens, index );
			}

			result.setMoveNumber(moveNumber);
			result.setIsWhiteToMove(whiteToPlay);

			if( nextPreToken!= null )
			{
				if( nextPreToken.getTokenId() != TokenId.NAG ) 
				{
					index._value--;
				}
			}
			else if( !hasToEmptyBuffer ) // if we do not have to empty the buffer, we need an extra token to be sure that no NAG or COMMENT follows for the next move.
			{
				result = null;
			}
		}

		return( result );
	}

	protected PreToken getPreTokenFromPreTokenString( String preTokenStr ) throws ChessParserException
	{
		PreToken result = null;

		if( preTokenStr != null )
		{
			int iniPosOfLine = ( _newLineMatcher != null ? _newLineMatcher.start() : 0 );

			TokenId tokenId = getTokenId( preTokenStr );

			result = new PreToken( tokenId, preTokenStr, _currentLineNumber, iniPosOfLine + _currentPosInLine - preTokenStr.length() + 1 );
			_currentPosInLine++;
		}

		return( result );
	}

	protected int getMoveNumber( LinkedList<PreToken> listOfPreTokens )
	{
		int result = -1;

		Iterator<PreToken> it = listOfPreTokens.iterator();
		while( (result==-1) && it.hasNext() )
		{
			PreToken preToken = it.next();
			if( preToken.getTokenId().equals( TokenId.NUMBER ) )
				result = Integer.parseInt( preToken.getString() );
		}
		return( result );
	}
	
	// 	TokenId[] case4 = { TokenId.MOVE, TokenId.NAG, TokenId.NUMBER, TokenId.DOT, TokenId.MOVE, TokenId.NAG, TokenId.MOVE, TokenId.NAG };
	protected Token createCase4Move( ResultOfCheckMoveTokenSequence rocmts, boolean hasToEmptyBuffer ) throws ChessParserException
	{
		Token result = null;

		LinkedList<PreToken> listOfPreTokens = rocmts.getResultListOfPreTokens();

		int moveNumber = getMoveNumber( listOfPreTokens );
		
		IntegerReference index = new IntegerReference(0);

		boolean whiteToPlay = false;
		MoveToken previousMove = getMoveTokenFromList( index, listOfPreTokens, moveNumber-1, whiteToPlay, hasToEmptyBuffer );

		index._value++;
		whiteToPlay = true;
		MoveToken firstMove = getMoveTokenFromList( index, listOfPreTokens, moveNumber, whiteToPlay, hasToEmptyBuffer );

		whiteToPlay = false;
		MoveToken secondMove = getMoveTokenFromList( index, listOfPreTokens, moveNumber, whiteToPlay, hasToEmptyBuffer );

		if( ( previousMove != null ) && ( firstMove != null ) && ( secondMove != null ) )
		{
			_bufferOfOutputTokens.add( previousMove );
			_bufferOfOutputTokens.add( firstMove );
			_bufferOfOutputTokens.add( secondMove );
			rocmts.setNumberOfInputItemsToErase(index);

			result = firstMove;
		}

		return( result );
	}

	protected Token createCase5Move( ResultOfCheckMoveTokenSequence rocmts, boolean hasToEmptyBuffer ) throws ChessParserException
	{
		Token result = null;

		LinkedList<PreToken> listOfPreTokens = rocmts.getResultListOfPreTokens();

		int moveNumber = getMoveNumber( listOfPreTokens );
		
		IntegerReference index = new IntegerReference(0);

		boolean whiteToPlay = false;
		MoveToken previousMove = getMoveTokenFromList( index, listOfPreTokens, moveNumber-1, whiteToPlay, hasToEmptyBuffer );

		index._value++;
		whiteToPlay = true;
		MoveToken firstMove = getMoveTokenFromList( index, listOfPreTokens, moveNumber, whiteToPlay, hasToEmptyBuffer );

		if( ( previousMove != null ) && ( firstMove != null )  )
		{
			_bufferOfOutputTokens.add( previousMove );
			_bufferOfOutputTokens.add( firstMove );
			rocmts.setNumberOfInputItemsToErase(index);

			result = firstMove;
		}

		return( result );
	}

	protected static ChessStringsConf getChessStrConf()
	{
		return( ChessStringsConf.instance() );
	}
}
