/* 
 * Copyright (C) 2017 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */
package com.frojasg1.chesspdfbrowser.model.chess.game;

import com.frojasg1.chesspdfbrowser.model.chess.board.ChessBoard;
import com.frojasg1.chesspdfbrowser.model.chess.board.ChessGamePosition;
import com.frojasg1.chesspdfbrowser.model.chess.configuration.ChessViewConfiguration;
import com.frojasg1.chesspdfbrowser.exception.ChessGamePositionException;
import com.frojasg1.chesspdfbrowser.exception.ChessModelException;
import com.frojasg1.chesspdfbrowser.exception.ChessMoveException;
import com.frojasg1.chesspdfbrowser.exception.ChessPieceCreationException;
import com.frojasg1.chesspdfbrowser.exception.ChessWriterException;
import com.frojasg1.chesspdfbrowser.model.chess.game.tree.ChessGameMove;
import com.frojasg1.chesspdfbrowser.io.writers.WrittenNodeInfo;
import com.frojasg1.chesspdfbrowser.io.writers.implementation.ChessGameViewWriter;
import com.frojasg1.chesspdfbrowser.model.chess.game.tree.MoveTreeNode;
import com.frojasg1.chesspdfbrowser.model.chess.game.tree.MoveTreeGame;
import java.util.ArrayList;
import java.util.List;

/**
 *
 * @author Usuario
 */
public class ChessGame 
{
	protected ChessBoard _chessBoard = null;

	protected ChessGameHeaderInfo _headerInfo = null;
	protected ChessGameTreeAdditionalInfo _chessGameTreeAdditionalInfo = null;
	protected MoveTreeGame _moveTreeGame = null;

	protected ChessViewConfiguration _chessViewConfiguration = null;

	protected MoveTreeNode _currentMove = null;

	protected List<ChessGameMove> _currentListOfMoves = null;

	protected int _currentPositionInListOfMoves = -1;
	
	public ChessGame( ChessViewConfiguration cvc ) throws ChessGamePositionException, ChessPieceCreationException
	{
		this( null, null, cvc );
	}

	public ChessGame( ChessGameHeaderInfo headerInfo,
						MoveTreeGame moveTreeGame,
						ChessViewConfiguration chessViewConfiguration
						) throws ChessGamePositionException, ChessPieceCreationException
	{
		if( headerInfo != null )
			_headerInfo = headerInfo;
		else
		{
			_headerInfo = new ChessGameHeaderInfo();
		}
		
		_chessViewConfiguration = chessViewConfiguration;

		if( moveTreeGame != null )
		{
			_moveTreeGame = moveTreeGame;
		}
		else
		{
			_moveTreeGame = new MoveTreeGame( ChessGamePosition.getInitialPosition() );
		}

		_currentMove = _moveTreeGame;
		_chessGameTreeAdditionalInfo = new ChessGameTreeAdditionalInfo( this );

		_chessBoard = new ChessBoard();
		
		if( _moveTreeGame.getInitialPosition() != null )
			_chessBoard.setPosition( _moveTreeGame.getInitialPosition() );

		_currentListOfMoves = new ArrayList<ChessGameMove>();
	}

	public ChessGameHeaderInfo getChessGameHeaderInfo()
	{
		return( _headerInfo );
	}

	public void setChessGameHeaderInfo( ChessGameHeaderInfo cghi )
	{
		_headerInfo = cghi;
	}

	public void setChessViewConfiguration( ChessViewConfiguration cvc )
	{
		_chessViewConfiguration = cvc;
	}

	public ChessViewConfiguration getChessViewConfiguration()
	{
		return( _chessViewConfiguration );
	}

	public ChessGamePosition getInitialPosition()
	{
		return( _moveTreeGame.getInitialPosition() );
	}

	public void setInitialPosition( ChessGamePosition cgp )
	{
		if( cgp != null )
		{
			ChessGamePosition previous = _moveTreeGame.getInitialPosition();
			try
			{
				_moveTreeGame.setInitialPosition( cgp );
				_chessBoard.setPosition( cgp );
				String fenString = cgp.getFenString();
				if( ! fenString.equals( ChessGamePosition.INITIAL_POSITION_FEN_STRING ) )
				{
					_headerInfo.put( ChessGameHeaderInfo.FEN_TAG, fenString );
					_headerInfo.put( ChessGameHeaderInfo.SETUP_TAG, "1" );
				}
				else
				{
					_headerInfo.remove( ChessGameHeaderInfo.FEN_TAG );
					_headerInfo.remove( ChessGameHeaderInfo.SETUP_TAG );
				}
			}
			catch( Throwable th )
			{
				th.printStackTrace();

				_moveTreeGame.setInitialPosition( previous );
				try
				{
					_chessBoard.setPosition( previous );
				}
				catch( Throwable th1 )
				{}
			}
		}
		else
			_moveTreeGame.setLevel(0);
	}

	public MoveTreeGame getMoveTreeGame()
	{
		return( _moveTreeGame );
	}

	public ChessBoard getChessBoard()
	{
		return( getInitialPosition() == null ? null : _chessBoard );
	}

	public ChessGameTreeAdditionalInfo getChessGameTreeAdditionalInfo()
	{
		return( _chessGameTreeAdditionalInfo );
	}

	public boolean isValid()
	{
		return( getChessBoard() != null );
	}

	public String updateAdditionalInfo( ) throws ChessGamePositionException, ChessPieceCreationException, ChessMoveException, ChessModelException, ChessWriterException
	{
		_chessGameTreeAdditionalInfo = new ChessGameTreeAdditionalInfo( this );

		_chessGameTreeAdditionalInfo.updateAdditionalInfo();

		return( _chessGameTreeAdditionalInfo.getMoveTreeString() );
	}

	public int getNumberOfMovesInTree()
	{
		int result = 0;

		if( _moveTreeGame != null ) result = _moveTreeGame.getNumberOfTotalElements();

		return( result );
	}

	public MoveTreeNode insertMoves( List<ChessGameMove> listOfMoves )
		throws ChessGamePositionException, ChessPieceCreationException, ChessMoveException, ChessModelException, ChessWriterException
	{
		MoveTreeNode result = _moveTreeGame.insertMoves(listOfMoves);
//		updateAdditionalInfo();

		return( result );
	}

	public boolean doMovesFromInitialPosition( List<ChessGameMove> listOfMoves ) throws ChessGamePositionException, ChessPieceCreationException, ChessMoveException
	{
		return( setCurrentMove( findNode( listOfMoves ) ) );
	}

	public MoveTreeNode getCurrentMove()
	{
		return( _currentMove );
	}
	
	public void start() throws ChessMoveException, ChessPieceCreationException, ChessGamePositionException
	{
		if( isValid() && ( _chessBoard != null ) )
		{
			if( setCurrentMove( _moveTreeGame ) )
				_currentPositionInListOfMoves = -1;
		}
	}

	public void forward() throws ChessMoveException, ChessPieceCreationException, ChessGamePositionException
	{
		if( isValid() && ( _chessBoard != null ) )
		{
			// if we are still inside the list of moves, we choose the next move. If not, we choose the next main line move
			if( _currentPositionInListOfMoves < (_currentListOfMoves.size()-1) )
			{
				MoveTreeNode mtn = _currentMove.findChild( _currentListOfMoves.get( _currentPositionInListOfMoves + 1 ) );
				if( mtn != null )	// mnt should never be null, but we check it.
				{
					if( setCurrentMove( mtn ) )
						_currentPositionInListOfMoves++;
				}
			}
			else if( _currentMove.getNumberOfChildren() > 0 )
			{
				if( setCurrentMove( _currentMove.getChild(0) ) )
					_currentPositionInListOfMoves++;
			}
		}
	}
	
	public void back() throws ChessMoveException, ChessPieceCreationException, ChessGamePositionException
	{
		if( isValid() && ( _chessBoard != null ) )
		{
			if( ( _currentMove.getParent() != null ) &&
				( _currentPositionInListOfMoves >= 0 ) )
			{
				if( setCurrentMove( _currentMove.getParent() ) )
					_currentPositionInListOfMoves--;
			}
		}
	}

	public void end() throws ChessMoveException, ChessPieceCreationException, ChessGamePositionException
	{
		if( isValid() && ( _chessBoard != null ) )
		{
			MoveTreeNode mtn = findNode( _currentListOfMoves );

			if( mtn != null )
			{
				if( setCurrentMove( mtn.getEndOfMainLine() ) )
					_currentPositionInListOfMoves = _currentMove.getLevel() - _moveTreeGame.getChild(0).getLevel();
			}
		}
	}

	public boolean setCurrentMove( MoveTreeNode mtn ) throws ChessMoveException, ChessPieceCreationException, ChessGamePositionException
	{
		boolean result = false;
		MoveTreeNode previous = _currentMove;
		try
		{
			setCurrentMove_internal( mtn );
			result = true;

			WrittenNodeInfo wni = null;
			if( mtn != null )
				wni = mtn.getAdditionalInfo();

			if( ( wni != null ) && ( wni.getIsIllegalMove() || wni.getIsAnyParentIllegalMove() ) )
			{
				_chessBoard.setCurrentNodeIsIllegal( true );
			}
		}
		catch( Throwable th )
		{
			setCurrentMove_internal( previous );	// if is has been some error, we reverse the operation.
			// currently is always allowed, even though the move is illegal or there is not an initial position.
		}
		return( result );
	}
/*
	protected void setCurrentMove_internal( MoveTreeNode mtn ) throws ChessMoveException, ChessPieceCreationException, ChessGamePositionException
	{
		if( ( mtn != null ) && ( _chessBoard != null ) && isValid() )
		{
			_currentMove = mtn;
			_chessBoard.doMovesFromInitialPosition( _moveTreeGame.getInitialPosition(), _currentMove.getGameMoveList() );
		}
	}
*/
	protected void setCurrentMove_internal( MoveTreeNode mtn ) throws ChessMoveException, ChessPieceCreationException, ChessGamePositionException
	{
		if( ( mtn != null ) &&
			( (mtn instanceof MoveTreeGame) || ( mtn.isParent( getMoveTreeGame() ) ) ) )
		{
			_currentMove = mtn;
			if(  ( _chessBoard != null ) && isValid() &&
				!( ( mtn.getAdditionalInfo() != null ) &&
					( mtn.getAdditionalInfo().getIsAnyParentIllegalMove() || 
						mtn.getAdditionalInfo().getIsIllegalMove() ) )
				)
			{
				_chessBoard.doMovesFromInitialPosition( _moveTreeGame.getInitialPosition(), _currentMove.getGameMoveList() );
			}
		}
	}

	public void setCurrentListOfMoves( List<ChessGameMove> listOfMoves ) throws ChessMoveException, ChessPieceCreationException, ChessGamePositionException
	{
//		if( isValid() && ( _chessBoard != null ) )
		if( _chessBoard != null )
		{
			MoveTreeNode mtn = findNode( listOfMoves );

			if( mtn != null )
			{
				_currentListOfMoves = listOfMoves;
				if( setCurrentMove( mtn ) )
					_currentPositionInListOfMoves = listOfMoves.size()-1;
			}
		}
	}

	public MoveTreeNode changeMoveDueToSelectionInView( int charPosition ) throws ChessMoveException, ChessPieceCreationException, ChessGamePositionException
	{
		MoveTreeNode result = null;
		WrittenNodeInfo nodeInfo = _chessGameTreeAdditionalInfo.getMoveAdditionalInfo(charPosition);
		if( nodeInfo != null )
		{
			result = nodeInfo.getMoveTreeNode();
			setCurrentMove( result );
			result = _currentMove;					// as if there was an error setting the new move.
		}
		return( result );
	}

	public MoveTreeNode findNode( List<ChessGameMove> listOfMoves )
	{
		MoveTreeNode result = null;

		if( ( _moveTreeGame != null ) && ( listOfMoves != null ) )
		{
			result = _moveTreeGame.findNode(listOfMoves);
		}

		return( result );
	}

	public String getMoveListString( List<ChessGameMove> listOfMoves ) throws ChessWriterException
	{
		ChessGameViewWriter pgnw = new ChessGameViewWriter( true );
		
		List<ChessGameMove> firstListOfMoves = ( listOfMoves != null ? listOfMoves : _chessBoard.getListOfMoves() );
		
		return( pgnw.getGameMovesStringForLine(this, firstListOfMoves ) + "\n" + 
				pgnw.getGameMovesStringForLine(this, _currentMove.getGameMoveList() ) + "\n" );
	}

	public String getMoveListString( MoveTreeNode mtn ) throws ChessWriterException
	{
		ChessGameViewWriter pgnw = new ChessGameViewWriter( true );

		String result = pgnw.getGameMovesStringForLine(this, mtn );
		return( result );
	}

	public int getNumberOfMovesOfMainLine()
	{
		int numberOfMoves = 0;

		if( ( _moveTreeGame != null ) && ( _moveTreeGame.getNumberOfChildren() > 0 ) )
		{
			MoveTreeNode mtn = _moveTreeGame.getEndOfMainLine();
			numberOfMoves = ( mtn.getLevel() - _moveTreeGame.getChild(0).getLevel() ) / 2 + 1;
		}

		return( numberOfMoves );
	}
}
