/* 
 * Copyright (C) 2017 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */
package com.frojasg1.general.desktop.dialogs.implementation;

import com.frojasg1.applications.common.configuration.application.BaseApplicationConfigurationInterface;
import java.awt.Component;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.HeadlessException;
import java.awt.Point;
import java.awt.Rectangle;
import java.io.File;
import java.util.Iterator;
import java.util.List;
import javax.swing.Icon;
import javax.swing.JDialog;
import javax.swing.JFileChooser;
import javax.swing.JMenu;
import javax.swing.JOptionPane;
import javax.swing.filechooser.FileFilter;

/**
 *
 * @author Usuario
 */
public class StaticDesktopDialogsWrapper
{
	public static final int OPEN = 0;
	public static final int SAVE = 1;
	
    protected StaticDesktopDialogsWrapper(){}
    
	static protected Font M_getNewFontForComponentFromApplicationFontSize( Component comp, float factor )
	{
		Font result = null;

		Font oldFont = comp.getFont();
		if( oldFont != null )
		{
			if( factor > 0.0F )
			{
				result = oldFont.deriveFont(factor * oldFont.getSize2D() );
			}
		}

		return( result );
	}
	
	static protected void M_changeFontToApplicationFontSize( Component comp, float factor )
	{
		Font newFont = M_getNewFontForComponentFromApplicationFontSize( comp, factor );
		if( newFont != null )
		{
			comp.setFont( newFont );
		}

		if( comp instanceof Container	)
		{
			Container contnr = (Container) comp;
			for( int ii=0; ii<contnr.getComponentCount(); ii++ )
			{
				M_changeFontToApplicationFontSize( contnr.getComponent(ii), factor );
			}

			if( contnr instanceof JMenu )
			{
				JMenu jmnu = (JMenu) contnr;
				for( int ii=0; ii<jmnu.getMenuComponentCount(); ii++ )
					M_changeFontToApplicationFontSize( jmnu.getMenuComponent( ii ), factor );
			}
		}
	}
    
	static public void M_changeFontToApplicationFontSize_forComponent( BaseApplicationConfigurationInterface conf, Component comp )
	{
		float factor = 0.0F;
		
		if( conf != null ) factor = conf.getFontSizeFactor();

		if( factor > 0.0F )
		{
			M_changeFontToApplicationFontSize( comp, factor );

			if( comp instanceof JFileChooser )
			{
				Dimension dim = comp.getPreferredSize();
				int width = (int) ( ( (float) dim.getWidth()) * factor );
				int height = (int) ( ( (float) dim.getHeight()) * factor );
				Dimension newDim = new Dimension( width, height);

				comp.setPreferredSize( newDim );
			}
		}
	}
	
	/**
	 * 
	 * @param parent
	 * @param message
	 * @param title
	 * @param messageType    an example for this parameter: JOptionPane.ERROR_MESSAGE
	 * @param conf 
	 */
	static public void showMessageDialog( Component parent, Object message, String title, int messageType, BaseApplicationConfigurationInterface conf )
	{
		JOptionPane option = new JOptionPane(	message, messageType );
		M_changeFontToApplicationFontSize_forComponent( conf, option );
		JDialog dialog = option.createDialog(parent, title );
		dialog.setAlwaysOnTop(true);
		dialog.setVisible(true);
	}

	/**
	 * 
	 * @param parentComponent
	 * @param message
	 * @param title
	 * @param optionType		an example for this parameter: JOptionPane.YES_NO_CANCEL_OPTION
	 * @param messageType		an example for this parameter: JOptionPane.QUESTION_MESSAGE
	 * @param icon
	 * @param options
	 * @param initialValue
	 * @param conf
	 * @return 
	 */
	static public int showOptionDialog( Component parentComponent, Object message, String title, int optionType,
										int messageType, Icon icon, Object[] options, Object initialValue,
										BaseApplicationConfigurationInterface conf )
	{
		JOptionPane option = new JOptionPane(	message, messageType );
		option.setIcon(icon);
		option.setOptionType(optionType);
		option.setOptions( options );
		option.setInitialValue( initialValue );
		M_changeFontToApplicationFontSize_forComponent( conf, option );
		JDialog dialog = option.createDialog(parentComponent, title );
		dialog.setAlwaysOnTop(true);
		dialog.setVisible(true);

		int result = -1;
		if( options != null )
		{
			for( int ii=0; (result==-1) && ii<options.length; ii++ )
				if( option.getValue() == options[ii] ) result = ii;
		}
		return( result );
	}
	
	/**
	 * 
	 * @param parent
	 * @param open_save_dialog	. This parameter has to contain either StaticDesktopDialogsWrapper.OPEN or
								StaticDesktopDialogsWrapper.SAVE
	 * @param conf
	 * @param fileFilterList
	 * @return 
	 */
	static public String showFileChooserDialog( Component parent,
													int open_save_dialog,
													BaseApplicationConfigurationInterface conf,
													List<FileFilter> fileFilterList,
													String defaultFileName )
	{
		String result = null;

		String path = "";

		if( conf != null ) path = conf.getLastDirectory();

		MyChooser chooser=new MyChooser(path);

		if( fileFilterList != null )
		{
			Iterator<FileFilter> it = fileFilterList.iterator();
			while( it.hasNext() )
			{
				chooser.addChoosableFileFilter(it.next());
			}
		}

        M_changeFontToApplicationFontSize_forComponent( conf, chooser );

		if( ( conf != null ) && ( conf.getLastFileChooserBounds() != null ) )
		{
			chooser.setDialogBounds( conf.getLastFileChooserBounds() );
		}

		int returnVal = -1;

		if( defaultFileName != null )
			chooser.setSelectedFile( new File( defaultFileName ) );

		if( open_save_dialog == OPEN )		returnVal = chooser.showOpenDialog(parent);
		else if( open_save_dialog == SAVE )	returnVal = chooser.showSaveDialog(parent);

		if(returnVal == JFileChooser.APPROVE_OPTION)
		{
			path=chooser.getSelectedFile().getParentFile().getAbsolutePath();
			result = chooser.getSelectedFile().toString();

			if( conf != null ) conf.setLastDirectory(path );
		}

		if( conf != null )
		{
			Rectangle newBounds = chooser.getDialogBounds();
			conf.setLastFileChooserBounds( newBounds );
		}

		return( result );
	}

	public static Point getCenteredLocationForComponent( Component comp )
	{
		int width = java.awt.Toolkit.getDefaultToolkit().getScreenSize().width;
		int height = java.awt.Toolkit.getDefaultToolkit().getScreenSize().height;
		Point result = new Point( width/2 - comp.getWidth()/2, height/2 - comp.getHeight()/2 );
		
		return( result );
	}

	static class MyChooser extends JFileChooser
	{
		JDialog _jDialog = null;

		Rectangle _initialBounds = null;

		public MyChooser( String path )
		{
			super( path );
		}

		@Override
		protected JDialog createDialog(Component parent) throws HeadlessException
		{
			_jDialog = super.createDialog(parent);

			if( _initialBounds != null )
				_jDialog.setBounds( _initialBounds );

			_jDialog.setAlwaysOnTop(true);

			return _jDialog;
		}

		public void setDialogBounds( Rectangle bounds )
		{
			_initialBounds = bounds;
		}

		public Rectangle getDialogBounds()
		{
			Rectangle result = null;
			if( _jDialog != null )
			{
				result = _jDialog.getBounds();
			}
			return( result );
		}
	}
}
